import functools
import shlex

from leapp.libraries.stdlib import CalledProcessError, run
from leapp.models import PCIDevices, PCIDevice


def aslist(f):
    ''' Decorator used to convert generator to list '''
    @functools.wraps(f)
    def inner(*args, **kwargs):
        return list(f(*args, **kwargs))
    return inner


def get_from_list(l, idx, default=''):
    ''' Get item at index from list or return a default '''
    try:
        ret = l[idx]
    except IndexError:
        ret = default

    return ret


@aslist
def parse_pci_devices(devices):
    ''' Parse lspci output and return a list of PCI devices '''
    for d in devices:
        raw = shlex.split(d)

        params = [r for r in raw if not r.startswith('-')]
        optionals = [r for r in raw if r.startswith('-')]

        rev = ''
        progif = ''
        for o in optionals:
            if o.startswith('-r'):
                rev = o.lstrip('-r')

            if o.startswith('-p'):
                progif = o.lstrip('-p')

        yield PCIDevice(
            slot=get_from_list(params, 0),
            dev_cls=get_from_list(params, 1),
            vendor=get_from_list(params, 2),
            name=get_from_list(params, 3),
            subsystem_vendor=get_from_list(params, 4),
            subsystem_name=get_from_list(params, 5),
            rev=rev,
            progif=progif
        )


def produce_pci_devices(producer, devices):
    ''' Produce a Leapp message with all PCI devices '''
    producer(PCIDevices(devices=devices))


def scan_pci_devices(producer):
    ''' Scan system PCI Devices '''
    try:
        output = run(['lspci', '-mm'], split=True)['stdout']
    except CalledProcessError:
        output = []

    devices = parse_pci_devices(output)
    produce_pci_devices(producer, devices)



from leapp.actors import Actor
from leapp.libraries.common import rhsm
from leapp.models import Report, SourceRHSMInfo
from leapp.tags import IPUWorkflowTag, ChecksPhaseTag
from leapp.libraries.common.reporting import report_with_remediation


class CheckRedHatSubscriptionManagerSKU(Actor):
    """
    Ensure the system is subscribed to the subscription manager

    This actor verifies that the system is correctly subscribed to via the Red Hat Subscription Manager and
    has attached SKUs. The actor will inhibit the upgrade if there are none.
    """

    name = 'check_rhsmsku'
    consumes = (SourceRHSMInfo,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        if not rhsm.skip_rhsm():
            for info in self.consume(SourceRHSMInfo):
                if not info.attached_skus:
                    report_with_remediation(
                        title='The system is not registered or subscribed.',
                        summary='The system has to be registered and subscribed to be able to proceed the upgrade.',
                        remediation=(
                            'Register your system with the subscription-manager tool and attach it to proper SKUs'
                            ' to be able to proceed the upgrade.'),
                        severity='high',
                        flags=['inhibitor']
                    )



from leapp.actors import Actor
from leapp.models import ActiveKernelModulesFacts
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_with_remediation


class CheckBtrfs(Actor):
    """
    Check if Btrfs filesystem is in use. If yes, inhibit the upgrade process.

    Btrfs filesystem was introduced as Technology Preview with initial releases of RHEL 6 and 7. It
    was deprecated on versions 6.6 and 7.4 and will not be present in next major version.
    """

    name = 'check_btrfs'
    consumes = (ActiveKernelModulesFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for fact in self.consume(ActiveKernelModulesFacts):
            for active_module in fact.kernel_modules:
                if active_module.filename == 'btrfs':
                    report_with_remediation(
                        title='Btrfs removed in the next major version',
                        summary='The Btrfs file system was introduced as Technology Preview with the initial release '
                                'of Red Hat Enterprise Linux 6 and Red Hat Enterprise Linux 7. As of versions 6.6 and '
                                '7.4 this technology has been deprecated and will be removed in next major version.',
                        remediation='Please consider migrating your Btrfs mount point(s) to a different filesystem '
                                    'before next upgrade attempt. If no Btrfs filesystem is in use, please unload '
                                    'btrfs kernel module running "# rmmod btrfs".',
                        severity='high',
                        flags=['inhibitor'])
                    break



from leapp.actors import Actor
from leapp.libraries.actor import library
from leapp.libraries.common.reporting import report_generic
from leapp.models import SendmailMigrationDecision
from leapp.reporting import Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateSendmail(Actor):
    """
    Migrate sendmail configuration files.
    """

    name = 'migrate_sendmail'
    consumes = (SendmailMigrationDecision,)
    produces = (Report,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        decision = next(self.consume(SendmailMigrationDecision), None)
        if not decision or not decision.migrate_files:
            return

        for f in decision.migrate_files:
            library.migrate_file(f)
        list_separator_fmt = '\n    - '
        report_generic(
            title='sendmail configuration files migrated',
            summary='Uncompressed IPv6 addresses in {}'.format(list_separator_fmt.join(decision.migrate_files)),
            severity='low')



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class OpenSshPermitRootLogin(Model):
    topic = SystemInfoTopic

    value = fields.StringEnum(['yes', 'prohibit-password',
                               'forced-commands-only', 'no'])
    in_match = fields.Nullable(fields.List(fields.String()))


class OpenSshConfig(Model):
    """
    OpenSSH server configuration.

    This model contains the first effective configuration option specified
    in the configuration file or a list of all the options specified
    in all the conditional blocks used throughout the file.
    """
    topic = SystemInfoTopic

    permit_root_login = fields.List(fields.Model(OpenSshPermitRootLogin))
    use_privilege_separation = fields.Nullable(fields.StringEnum(['sandbox',
                                                                  'yes',
                                                                  'no']))
    protocol = fields.Nullable(fields.String())
    ciphers = fields.Nullable(fields.String())
    macs = fields.Nullable(fields.String())

    modified = fields.Nullable(fields.Boolean())



from leapp.actors import Actor
from leapp.libraries.actor import library
from leapp.models import InstalledRedHatSignedRPM, VsftpdFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class VsftpdConfigRead(Actor):
    '''
    Reads vsftpd configuration files (/etc/vsftpd/*.conf) and extracts necessary information.
    '''

    name = 'vsftpd_config_read'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (VsftpdFacts,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        installed_rpm_facts = next(self.consume(InstalledRedHatSignedRPM))
        if library.is_processable(installed_rpm_facts):
            self.produce(library.get_vsftpd_facts())



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class InitrdIncludes(Model):
    """
    List of files (cannonical filesystem paths) to include in RHEL-8 initrd
    """
    topic = SystemInfoTopic

    files = fields.List(fields.String())



from leapp.actors import Actor
from leapp.models import InstalledTargetKernelVersion, KernelCmdlineArg
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag
from leapp.libraries.actor import kernelcmdlineconfig


class KernelCmdlineConfig(Actor):
    """
    Append extra arguments to RHEL-8 kernel command line
    """

    name = 'kernelcmdlineconfig'
    consumes = (KernelCmdlineArg, InstalledTargetKernelVersion)
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        kernelcmdlineconfig.process()



import functools
import sys

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import mounting
from leapp.libraries.stdlib import STDOUT, CalledProcessError, api, config, run


def makedirs(path, mode=0o777, exists_ok=True):
    mounting._makedirs(path=path, mode=mode, exists_ok=exists_ok)


def apply_yum_workaround(context=None):
    """
    Applies a workaround on the system to allow the upgrade to succeed for yum/dnf.
    """
    yum_script_path = api.get_tool_path('handleyumconfig')
    if not yum_script_path:
        raise StopActorExecutionError(
            message='Failed to find mandatory script to apply',
            details=reinstall_leapp_repository_hint()
        )
    cmd = ['/bin/bash', '-c', yum_script_path]

    try:
        context = context or mounting.NotIsolatedActions(base_dir='/')
        context.call(cmd)
    except OSError as e:
        raise StopActorExecutionError(
            message='Failed to exceute script to apply yum adjustment. Message: {}'.format(str(e))
        )
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Failed to apply yum adjustment. Message: {}'.format(str(e))
        )


def logging_handler(fd_info, buf):
    """
    Custom log handler to always show stdout to console and stderr only in DEBUG mode
    """
    (_unused, fd_type) = fd_info

    if fd_type == STDOUT:
        sys.stdout.write(buf)
    else:
        if config.is_debug():
            sys.stderr.write(buf)


def reinstall_leapp_repository_hint():
    """
    Convenience function for creating a detail for StopActorExecutionError with a hint to reinstall the
    leapp-repository package
    """
    return {
        'hint': 'Try to reinstall the `leapp-repository` package.'
    }


def report_and_ignore_shutil_rmtree_error(func, path, exc_info):
    """
    Helper function for shutil.rmtree to only report errors but don't fail.
    """
    api.current_logger().warn(
        'While trying to remove directories: %s failed at %s with an exception %s message: %s',
        func.__name__, path, exc_info[0].__name__, exc_info[1]
    )


def call_with_oserror_handled(cmd):
    """
    Perform run with already handled OSError for some convenience.
    """
    try:
        run(cmd)
    except OSError as e:
        if cmd:
            raise StopActorExecutionError(
                message=str(e),
                details={
                    'hint': 'Please ensure that {} is installed and executable.'.format(cmd[0])
                }
            )
        else:
            raise StopActorExecutionError(
                message='Failed to execute command {} with: {}'.format(''.join(cmd), str(e))
            )


def call_with_failure_hint(cmd, hint):
    """
    Perform `run` which handles OSError through call_with_oserror_handled and transforms CalledProcessError to a
    StopActorExecutionError with a hint given by the caller.
    """
    try:
        call_with_oserror_handled(cmd)
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Failed to execute command `{}`. Error: {}'.format(' '.join(cmd), str(e)),
            details={hint: hint}
        )


def clean_guard(cleanup_function):
    """
    Decorator to handle any exception going through and running cleanup tasks through the given cleanup_function
    parameter.
    """
    def clean_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:  # Broad exception handler to handle all cases but rethrows
                api.current_logger().debug('Clean guard caught an exception - Calling cleanup function.')
                try:
                    cleanup_function(*args, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    # Broad exception handler to handle all cases however, swallowed, to avoid loosing the original
                    # error. Logging for debuggability.
                    api.current_logger().warn('Caught and swallowed an exception during cleanup.', exc_info=True)
                raise  # rethrow original exception
        return wrapper
    return clean_wrapper



import shutil
import os

from leapp.actors import Actor
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_generic


class CreateSystemdResumeService(Actor):
    """
    Add a systemd service to launch Leapp.

    Create a systemd service which will resume Leapp upgrade after the first reboot.
    """

    name = 'create_systemd_service'
    consumes = ()
    produces = (Report,)
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        service_name = 'leapp_resume.service'
        service_templ_fpath = self.get_file_path(service_name)
        shutil.copyfile(service_templ_fpath, os.path.join(
            '/etc/systemd/system/', service_name))

        service_path = '/etc/systemd/system/{}'.format(service_name)
        symlink_path = '/etc/systemd/system/default.target.wants/{}'.format(service_name)

        try:
            os.symlink(service_path, symlink_path)
        except OSError as e:
            self.report_error('Could not create a symlink to enable {}'.format(service_name),
                              details=str(e))
            return

        report_generic(
            title='Leapp resume systemd service enabled',
            summary='{} enabled as oneshot systemd service to resume Leapp execution '
                    'after reboot.'.format(service_name),
            severity='low')



import os

from leapp.utils.repository import find_repository_basedir
from leapp.repository.scan import find_and_scan_repositories


def pytest_sessionstart(session):
    actor_path = os.environ.get('LEAPP_TESTED_ACTOR', None)
    library_path = os.environ.get('LEAPP_TESTED_LIBRARY', None)

    if actor_path:
        repo = find_and_scan_repositories(find_repository_basedir(actor_path), include_locals=True)
        repo.load()

        actor = None
        # find which actor is being tested
        for a in repo.actors:
            if a.full_path == actor_path.rstrip('/'):
                actor = a
                break

        if not actor:
            return

        # load actor context so libraries can be imported on module level
        session.leapp_repository = repo
        session.actor_context = actor.injected_context()
        session.actor_context.__enter__()
    elif library_path:
        repo = find_and_scan_repositories(find_repository_basedir(library_path), include_locals=True)
        repo.load()
        os.chdir(library_path)



from leapp.libraries.actor import setrelease
from leapp.libraries.common import mounting, rhsm
from leapp.libraries.stdlib import api
from leapp.models import TargetRHSMInfo


def not_isolated_actions():
    commands_called = []

    class MockNotIsolatedActions(object):
        def __init__(self, base_dir=None):
            pass

        def call(self, cmd, **kwargs):
            commands_called.append((cmd, kwargs))
    return (commands_called, MockNotIsolatedActions)


def test_setrelease(monkeypatch):
    commands_called, klass = not_isolated_actions()
    monkeypatch.setattr(mounting, 'NotIsolatedActions', klass)
    monkeypatch.setattr(api, 'consume', lambda *x: (x for x in (TargetRHSMInfo(release='6.6.6'),)))
    setrelease.process()
    assert commands_called and len(commands_called) == 1
    assert commands_called[0][0][-1] == '6.6.6'


def test_setrelease_no_message(monkeypatch):
    commands_called, klass = not_isolated_actions()
    monkeypatch.setattr(mounting, 'NotIsolatedActions', klass)
    monkeypatch.setattr(api, 'consume', lambda *x: (x for x in ()))
    setrelease.process()
    assert not commands_called


def test_setrelease_skip_rhsm(monkeypatch):
    commands_called, klass = not_isolated_actions()
    monkeypatch.setenv('LEAPP_DEVEL_SKIP_RHSM', '1')
    # To make this work we need to re-apply the decorator, so it respects the environment variable
    monkeypatch.setattr(rhsm, 'set_release', rhsm.with_rhsm(rhsm.set_release))
    monkeypatch.setattr(mounting, 'NotIsolatedActions', klass)
    monkeypatch.setattr(api, 'consume', lambda *x: (x for x in (TargetRHSMInfo(release='6.6.6'),)))
    setrelease.process()
    assert not commands_called



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_with_remediation
from leapp.libraries.common.rpms import has_package
from leapp.models import InstalledRedHatSignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class PowerTop(Actor):
    """
    Check if PowerTOP is installed. If yes, write information about non-compatible changes.
    """

    name = 'powertop'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(InstalledRedHatSignedRPM, 'powertop'):
            report_with_remediation(
                title='PowerTOP compatibility options removed in the next major version',
                summary='The -d (dump) option which has been kept for RHEL backward compatibility has been '
                        'dropped.\n'
                        'The -h option which has been used for RHEL backward compatibility is no longer '
                        'alias for --html, but it\'s now an alias for --help to follow the upstream.\n'
                        'The -u option which has been used for RHEL backward compatibility as an alias for '
                        '--help has been dropped.\n',
                remediation='Please remove the dropped options from your scripts.',
                severity='low')



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class PCIDevice(Model):
    topic = SystemInfoTopic

    slot = fields.String()
    dev_cls = fields.String()
    vendor = fields.String()
    name = fields.String()
    subsystem_vendor = fields.String()
    subsystem_name = fields.String()
    rev = fields.Nullable(fields.String())
    progif = fields.Nullable(fields.String())


class PCIDevices(Model):
    topic = SystemInfoTopic

    devices = fields.List(fields.Model(PCIDevice))



import os

from leapp.libraries.common import reporting
from leapp.models import EnvVar, OSRelease


ENV_IGNORE = ('LEAPP_CURRENT_PHASE', 'LEAPP_CURRENT_ACTOR', 'LEAPP_VERBOSE',
              'LEAPP_DEBUG')


def get_env_vars():
    """
    Gather LEAPP_DEVEL environment variables and provide them as messages to be
    available after reboot.
    """
    return [EnvVar(name=k, value=v) for (k, v) in os.environ.items() if k.startswith('LEAPP_') and k not in ENV_IGNORE]


def get_os_release(path):
    """Retrieve data about System OS release from provided file."""
    try:
        with open(path) as f:
            data = dict(l.strip().split('=', 1) for l in f.readlines() if '=' in l)
            return OSRelease(
                release_id=data.get('ID', '').strip('"'),
                name=data.get('NAME', '').strip('"'),
                pretty_name=data.get('PRETTY_NAME', '').strip('"'),
                version=data.get('VERSION', '').strip('"'),
                version_id=data.get('VERSION_ID', '').strip('"'),
                variant=data.get('VARIANT', '').strip('"') or None,
                variant_id=data.get('VARIANT_ID', '').strip('"') or None
            )
    except IOError as e:
        reporting.report_generic(
            title='Error while collecting system OS facts',
            summary=str(e),
            severity='high',
            flags=['inhibitor'])
        return None



from leapp.actors import Actor
from leapp.libraries.common import utils
from leapp.tags import IPUWorkflowTag, PreparationPhaseTag


class Prepareyumconfig(Actor):
    """
    Handle migration of the yum configuration files.

    RPM cannot handle replacement of directories by symlinks by default
    without the %pretrans scriptlet. As yum package is packaged wrong,
    we have to workround that by migration of the yum configuration files
    before the rpm transaction is processed.
    """

    name = 'prepareyumconfig'
    consumes = ()
    produces = ()
    tags = (IPUWorkflowTag, PreparationPhaseTag)

    def process(self):
        utils.apply_yum_workaround()



import re

import pyudev

from leapp.actors import Actor
from leapp.models import PersistentNetNamesFacts, KernelCmdlineArg
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class Biosdevname(Actor):
    """
    Enable biosdevname on RHEL8 if all interfaces on RHEL7 use biosdevname naming scheme and if machine vendor is DELL
    """

    name = 'biosdevname'
    consumes = (PersistentNetNamesFacts,)
    produces = (KernelCmdlineArg,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def is_biosdevname_disabled(self):
        with open('/proc/cmdline') as cmdline:
            if 'biosdevname=0' in cmdline.read():
                return True

        return False

    def is_vendor_dell(self):
        context = pyudev.Context()

        # There should be only one dmi/id device
        for dev in pyudev.Enumerator(context).match_subsystem('dmi').match_sys_name('id'):
            vendor = dev.attributes.get('sys_vendor')
            return re.search('Dell.*', str(vendor)) is not None
        return False

    def all_interfaces_biosdevname(self, interfaces):
        # Biosdevname supports two naming schemes
        emx = re.compile('em[0-9]+')
        pxpy = re.compile('p[0-9]+p[0-9]+')

        for i in interfaces:
            if emx.match(i.name) is None and pxpy.match(i.name) is None:
                return False

        return True

    def enable_biosdevname(self):
        self.log.info("Biosdevname naming scheme in use, explicitely enabling biosdevname on RHEL-8")
        self.produce(KernelCmdlineArg(**{'key': 'biosdevname', 'value': '1'}))

    def process(self):
        interfaces = next(self.consume(PersistentNetNamesFacts)).interfaces

        if self.is_biosdevname_disabled():
            return

        if self.is_vendor_dell() and self.all_interfaces_biosdevname(interfaces):
            self.enable_biosdevname()



from leapp.models import Report, TargetRHSMInfo


def test_report_target_version(current_actor_context):
    current_actor_context.feed(TargetRHSMInfo(release='6.6.6'))
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 1
    assert '6.6.6' in reports[0].detail.get('summary', '')
    assert '6.6.6' in reports[0].title


def test_report_target_version_notset(current_actor_context):
    current_actor_context.feed(TargetRHSMInfo(release=''))
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert not reports


def test_report_target_version_nomessage(current_actor_context):
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert not reports



import os.path

import pytest

from leapp.exceptions import StopActorExecution
from leapp.libraries.actor import library
from leapp.libraries.actor.library import (Event,
                                           add_output_pkgs_to_transaction_conf,
                                           filter_out_pkgs_in_blacklisted_repos,
                                           get_events,
                                           get_events_for_installed_pkgs_only,
                                           map_repositories, parse_action,
                                           parse_entry, parse_packageset,
                                           parse_pes_events_file,
                                           process_events,
                                           report_skipped_packages)
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import produce_mocked, report_generic_mocked
from leapp.libraries.stdlib import api
from leapp.models import RpmTransactionTasks


class show_message_mocked(object):
    def __init__(self):
        self.called = 0
        self.msg = None

    def __call__(self, msg):
        self.called += 1
        self.msg = msg


class get_repos_blacklisted_mocked(object):
    def __init__(self, blacklisted):
        self.blacklisted = blacklisted

    def __call__(self):
        return self.blacklisted


def test_parse_action(current_actor_context):
    assert parse_action(0) == 'Present'
    assert parse_action(1) == 'Removed'
    assert parse_action(2) == 'Deprecated'
    assert parse_action(3) == 'Replaced'
    assert parse_action(4) == 'Split'
    assert parse_action(5) == 'Merged'
    assert parse_action(6) == 'Moved'
    assert parse_action(7) == 'Renamed'

    with pytest.raises(ValueError):
        parse_action(-1)
        parse_action(8)


def test_parse_packageset(current_actor_context):
    pkgset = {'package': [{'name': 'pkg1', 'repository': 'Repo'}]}

    assert parse_packageset(pkgset) == {'pkg1': 'repo'}

    assert parse_packageset({}) == {}
    assert parse_packageset({'setid': 0}) == {}


def test_parse_entry(current_actor_context):
    entry = {
        'action': 4,
        'in_packageset': {
            'package': [{'name': 'original', 'repository': 'repo'}]},
        'out_packageset': {
            'package': [
                {'name': 'split01', 'repository': 'repo'},
                {'name': 'split02', 'repository': 'repo'}]}}

    event = parse_entry(entry)
    assert event.action == 'Split'
    assert event.in_pkgs == {'original': 'repo'}
    assert event.out_pkgs == {'split01': 'repo', 'split02': 'repo'}

    entry = {
        'action': 1,
        'in_packageset': {
            'package': [{'name': 'removed', 'repository': 'repo'}]}}

    event = parse_entry(entry)
    assert event.action == 'Removed'
    assert event.in_pkgs == {'removed': 'repo'}
    assert event.out_pkgs == {}


def test_parse_pes_events_file(current_actor_context):
    events = parse_pes_events_file('files/tests/sample01.json')
    assert len(events) == 2
    assert events[0].action == 'Split'
    assert events[0].in_pkgs == {'original': 'repo'}
    assert events[0].out_pkgs == {'split01': 'repo', 'split02': 'repo'}
    assert events[1].action == 'Removed'
    assert events[1].in_pkgs == {'removed': 'repo'}
    assert events[1].out_pkgs == {}


def test_get_events_for_installed_pkgs_only(monkeypatch):
    events = [
        Event('Split', {'original': 'repo'}, {'split01': 'repo', 'split02': 'repo'}),
        Event('Removed', {'removed': 'repo'}, {})]
    filtered = get_events_for_installed_pkgs_only(events, {'original'})

    assert len(filtered) == 1
    assert filtered[0].action == 'Split'
    assert filtered[0].in_pkgs == {'original': 'repo'}
    assert filtered[0].out_pkgs == {'split01': 'repo', 'split02': 'repo'}


def test_report_skipped_packages(monkeypatch):
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'show_message', show_message_mocked())
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setenv('LEAPP_VERBOSE', '1')
    report_skipped_packages('packages will not be installed:', ['skipped01', 'skipped02'])

    message = '2 packages will not be installed:\n- skipped01\n- skipped02'
    assert api.show_message.called == 1
    assert api.show_message.msg == message
    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'Packages will not be installed'
    assert reporting.report_generic.report_fields['summary'] == message


def test_report_skipped_packages_no_verbose_mode(monkeypatch):
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'show_message', show_message_mocked())
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setenv('LEAPP_VERBOSE', '0')
    report_skipped_packages('packages will not be installed:', ['skipped01', 'skipped02'])

    message = '2 packages will not be installed:\n- skipped01\n- skipped02'
    assert api.show_message.called == 0
    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'Packages will not be installed'
    assert reporting.report_generic.report_fields['summary'] == message


def test_filter_out_pkgs_in_blacklisted_repos(monkeypatch):
    monkeypatch.setattr(api, 'show_message', show_message_mocked())
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setattr(library, 'get_repositories_blacklisted', get_repos_blacklisted_mocked(set(['blacklisted'])))
    monkeypatch.setenv('LEAPP_VERBOSE', '1')

    to_install = {
        'pkg01': 'repo01',
        'pkg02': 'repo02',
        'skipped01': 'blacklisted',
        'skipped02': 'blacklisted'}
    filter_out_pkgs_in_blacklisted_repos(to_install)

    msgs = [
        '2 packages will not be installed due to blacklisted repositories:',
        '- skipped01',
        '- skipped02']
    assert api.show_message.called == 1
    assert api.show_message.msg == '\n'.join(msgs)
    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['summary'] == '\n'.join(msgs)
    assert reporting.report_generic.report_fields['title'] == 'Packages will not be installed'

    assert to_install == {'pkg01': 'repo01', 'pkg02': 'repo02'}


def test_map_repositories(monkeypatch):
    monkeypatch.setattr(api, 'show_message', show_message_mocked())
    monkeypatch.setattr(library, 'REPOSITORIES_MAPPING', {'repo': 'mapped'})
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setenv('LEAPP_VERBOSE', '1')

    to_install = {
        'pkg01': 'repo',
        'pkg02': 'repo',
        'skipped01': 'not_mapped',
        'skipped02': 'not_mapped'}
    map_repositories(to_install)

    msgs = [
        '2 packages will not be installed due to not mapped repositories:',
        '- skipped01',
        '- skipped02']
    assert api.show_message.called == 1
    assert api.show_message.msg == '\n'.join(msgs)
    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'Packages will not be installed'
    assert reporting.report_generic.report_fields['summary'] == '\n'.join(msgs)

    assert to_install == {'pkg01': 'mapped', 'pkg02': 'mapped'}


def test_process_events(monkeypatch):
    monkeypatch.setattr(library, 'REPOSITORIES_MAPPING', {'repo': 'mapped'})
    monkeypatch.setattr(library, 'get_repositories_blacklisted', get_repos_blacklisted_mocked(set()))

    events = [
        Event('Split', {'original': 'repo'}, {'split01': 'repo', 'split02': 'repo'}),
        Event('Removed', {'removed': 'repo'}, {})]
    to_install, to_remove = process_events(events, {})

    assert len(to_install) == 2
    assert to_install == {'split02': 'mapped', 'split01': 'mapped'}
    assert len(to_remove) == 2
    assert to_remove == {'removed': 'repo', 'original': 'repo'}


def test_get_events(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())

    with pytest.raises(StopActorExecution):
        get_events('files/tests/sample02.json')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']

    reporting.report_generic.called = 0
    reporting.report_generic.model_instances = []
    with pytest.raises(StopActorExecution):
        get_events('files/tests/sample03.json')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']


def test_pes_data_not_found(monkeypatch):
    def file_not_exists(_filepath):
        return False
    monkeypatch.setattr(os.path, 'isfile', file_not_exists)
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    with pytest.raises(StopActorExecution):
        get_events('/etc/leapp/pes-data.json')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']


def test_add_output_pkgs_to_transaction_conf():
    events = [
        Event('Split', {'split_in': 'repo'}, {'split_out1': 'repo', 'split_out2': 'repo'}),
        Event('Merged', {'merged_in1': 'repo', 'merged_in2': 'repo'}, {'merged_out': 'repo'}),
        Event('Renamed', {'renamed_in': 'repo'}, {'renamed_out': 'repo'}),
        Event('Replaced', {'replaced_in': 'repo'}, {'replaced_out': 'repo'}),
    ]

    conf_empty = RpmTransactionTasks()
    add_output_pkgs_to_transaction_conf(conf_empty, events)
    assert conf_empty.to_remove == []

    conf_split = RpmTransactionTasks(to_remove=['split_in'])
    add_output_pkgs_to_transaction_conf(conf_split, events)
    assert sorted(conf_split.to_remove) == ['split_in', 'split_out1', 'split_out2']

    conf_merged_incomplete = RpmTransactionTasks(to_remove=['merged_in1'])
    add_output_pkgs_to_transaction_conf(conf_merged_incomplete, events)
    assert conf_merged_incomplete.to_remove == ['merged_in1']

    conf_merged = RpmTransactionTasks(to_remove=['merged_in1', 'merged_in2'])
    add_output_pkgs_to_transaction_conf(conf_merged, events)
    assert sorted(conf_merged.to_remove) == ['merged_in1', 'merged_in2', 'merged_out']

    conf_renamed = RpmTransactionTasks(to_remove=['renamed_in'])
    add_output_pkgs_to_transaction_conf(conf_renamed, events)
    assert sorted(conf_renamed.to_remove) == ['renamed_in', 'renamed_out']

    conf_replaced = RpmTransactionTasks(to_remove=['replaced_in'])
    add_output_pkgs_to_transaction_conf(conf_replaced, events)
    assert sorted(conf_replaced.to_remove) == ['replaced_in', 'replaced_out']



from leapp.snactor.fixture import current_actor_context
from leapp.models import (Report, SELinuxFacts, SelinuxPermissiveDecision,
                          SelinuxRelabelDecision)


# FIXME: fix the file properly regarding the fix of the issue:
# # https://github.com/oamg/leapp-repository/issues/20
def create_selinuxfacts(static_mode, enabled, policy='targeted', mls_enabled=True):
    runtime_mode = static_mode if static_mode != 'disabled' else None

    return SELinuxFacts(
        runtime_mode=runtime_mode,
        static_mode=static_mode,
        enabled=enabled,
        policy=policy,
        mls_enabled=mls_enabled
    )


def test_actor_schedule_relabelling(current_actor_context):
    facts = [create_selinuxfacts(static_mode='permissive', enabled=True),
             create_selinuxfacts(static_mode='enforcing', enabled=True)]

    for fact in facts:
        current_actor_context.feed(fact)
        current_actor_context.run()
        assert current_actor_context.consume(Report)
        assert current_actor_context.consume(SelinuxRelabelDecision)[0].set_relabel


def test_actor_set_permissive(current_actor_context):
    relabel = create_selinuxfacts(static_mode='enforcing', enabled=True)

    current_actor_context.feed(relabel)
    current_actor_context.run()
    assert current_actor_context.consume(Report)
    assert current_actor_context.consume(SelinuxPermissiveDecision)[0].set_permissive


def test_actor_selinux_disabled(current_actor_context):
    disabled = create_selinuxfacts(enabled=False, static_mode='disabled')

    current_actor_context.feed(disabled)
    current_actor_context.run()
    assert not current_actor_context.consume(SelinuxRelabelDecision)
    assert not current_actor_context.consume(SelinuxPermissiveDecision)



import errno

from leapp.models import OpenSshConfig, OpenSshPermitRootLogin
from leapp.libraries.stdlib import run, api


CONFIG = '/etc/ssh/sshd_config'


def line_empty(line):
    return len(line) == 0 or line.startswith('\n') or line.startswith('#')


def parse_config(config):
    """Parse OpenSSH server configuration or the output of sshd test option."""

    # RHEL7 defaults
    ret = OpenSshConfig(
        permit_root_login=[],
    )

    in_match = None
    for line in config:
        line = line.strip()
        if line_empty(line):
            continue

        el = line.split()
        if len(el) < 2:
            continue
        value = el[1]
        if el[0].lower() == 'match':
            in_match = el[1:]
            continue

        if el[0].lower() == 'permitrootlogin':
            # convert deprecated alias
            if value == "without-password":
                value = "prohibit-password"
            v = OpenSshPermitRootLogin(value=value, in_match=in_match)
            ret.permit_root_login.append(v)

        elif el[0].lower() == 'useprivilegeseparation':
            # Record only first occurence, which is effective
            if not ret.use_privilege_separation:
                ret.use_privilege_separation = value

        elif el[0].lower() == 'protocol':
            # Record only first occurence, which is effective
            if not ret.protocol:
                ret.protocol = value

        elif el[0].lower() == 'ciphers':
            # Record only first occurence, which is effective
            if not ret.ciphers:
                ret.ciphers = value

        elif el[0].lower() == 'macs':
            # Record only first occurence, which is effective
            if not ret.macs:
                ret.macs = value

    return ret


def produce_config(producer, config):
    """Produce a Leapp message with all interesting OpenSSH configuration options."""

    producer(config)


def read_sshd_config():
    """Read the actual sshd configuration file."""
    try:
        with open(CONFIG, 'r') as fd:
            return fd.readlines()
    except IOError as err:
        if err.errno != errno.ENOENT:
            error = 'Failed to read sshd_config: {}'.format(str(err))
            api.current_logger().error(error)
        return []


def read_rpm_modifications():
    """Asks RPM database whether the configuration file was modified."""

    try:
        return run(['rpm', '-Vf', CONFIG], split=True, checked=False)['stdout']
    except OSError as err:
        error = 'Failed to check the modification status of the {}: {}' \
                ''.format(CONFIG, str(err))
        api.current_logger().error(error)
        return []


def parse_config_modification(data):
    """Handle the output of rpm verify to figure out configuration file was modified."""

    # First assume it is not modified -- empty data says it is not modified
    modified = False
    for line in data:
        parts = line.split(' ')
        # The last part of the line is the actual file we care for
        if parts[-1] == CONFIG:
            # First part contains information, if the size and digest differ
            if '5' in parts[0] or 'S' in parts[0]:
                modified = True
        # Ignore any other files lurking here

    return modified


def scan_sshd(producer):
    """Parse sshd_config configuration file to create OpenSshConfig message."""

    # direct access to configuration file
    output = read_sshd_config()
    config = parse_config(output)

    # find out whether the file was modified from the one shipped in rpm
    output = read_rpm_modifications()
    config.modified = parse_config_modification(output)

    produce_config(producer, config)



from leapp.models import RepositoriesBlacklisted, RepositoriesFacts, RepositoryFile, RepositoryData
from leapp.snactor.fixture import current_actor_context


def get_repo_data(repoids, enabled=True):
    data = []
    for repo in repoids:
        data.append(RepositoryData(repoid=repo, name=repo, enabled=enabled))
    return data


def get_repo_files(repoids, enabled=True):
    files = []
    for i, repo in enumerate(repoids):
        files.append(RepositoryFile(file='/etc/yum.d/sample{}.repo'.format(i),
                                    data=get_repo_data([repo])))
    return files


def get_repo_facts(repoids, enabled=True, multiple_files=False):
    if multiple_files:
        repos = get_repo_files(repoids, enabled=enabled)
    else:
        repos = [RepositoryFile(file='/etc/yum.d/sample.repo',
                                data=get_repo_data(repoids, enabled))]
    return RepositoriesFacts(repositories=repos)


def test_repositoriesblacklist_empty(current_actor_context):
    current_actor_context.feed()
    current_actor_context.run()
    assert current_actor_context.consume(RepositoriesBlacklisted)
    repoids = current_actor_context.consume(RepositoriesBlacklisted)[0].repoids
    assert 'codeready-builder-for-rhel-8-x86_64-rpms' in repoids


def test_repositoriesblacklist_no_optional(current_actor_context):
    repo_facts = get_repo_facts(['rhel-7-server-rpms', 'rhel-7-extras-rpms'])
    current_actor_context.feed(repo_facts)
    current_actor_context.run()
    assert current_actor_context.consume(RepositoriesBlacklisted)
    repoids = current_actor_context.consume(RepositoriesBlacklisted)[0].repoids
    assert 'codeready-builder-for-rhel-8-x86_64-rpms' in repoids


def test_repositoriesblacklist_disabled_optional(current_actor_context):
    repo_facts = get_repo_facts(['rhel-7-server-optional-rpms'], False)
    current_actor_context.feed(repo_facts)
    current_actor_context.run()
    assert current_actor_context.consume(RepositoriesBlacklisted)
    repoids = current_actor_context.consume(RepositoriesBlacklisted)[0].repoids
    assert 'codeready-builder-for-rhel-8-x86_64-rpms' in repoids


def test_repositoriesblacklist_optional(current_actor_context):
    repo_facts = get_repo_facts(['test', 'rhel-7-server-optional-rpms', 'rhel-7-server-rpms'])
    current_actor_context.feed(repo_facts)
    current_actor_context.run()
    assert not current_actor_context.consume(RepositoriesBlacklisted)


def test_repositoriesblacklist_optional_multiple_repo_files(current_actor_context):
    repo_facts = get_repo_facts(['test', 'rhel-7-server-optional-rpms', 'rhel-7-server-rpms'], True, True)
    current_actor_context.feed(repo_facts)
    current_actor_context.run()
    assert not current_actor_context.consume(RepositoriesBlacklisted)


def test_repositoriesblacklist_no_optional_multiple_repo_files(current_actor_context):
    repo_facts = get_repo_facts(['rhel-7-server-rpms', 'rhel-7-extras-rpms'], True, True)
    current_actor_context.feed(repo_facts)
    current_actor_context.run()
    assert current_actor_context.consume(RepositoriesBlacklisted)
    repoids = current_actor_context.consume(RepositoriesBlacklisted)[0].repoids
    assert 'codeready-builder-for-rhel-8-x86_64-rpms' in repoids



import mock

from leapp.libraries.common import rpms
from leapp.models import fields, InstalledRPM, InstalledRedHatSignedRPM, InstalledUnsignedRPM, Model, RPM
from leapp.snactor.fixture import current_actor_context


RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


class MockModel(Model):
    topic = RPM.topic
    list_field = fields.List(fields.Integer(), default=[42])
    list_field_nullable = fields.Nullable(fields.List(fields.String()))
    int_field = fields.Integer(default=42)


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(InstalledRedHatSignedRPM)


def test_actor_execution_with_signed_unsigned_data(current_actor_context):
    installed_rpm = [
        RPM(name='sample01', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
        RPM(name='sample02', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_OTHER_SIG_X'),
        RPM(name='sample03', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186'),
        RPM(name='sample04', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_OTHER_SIG_X'),
        RPM(name='sample05', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 938a80caf21541eb'),
        RPM(name='sample06', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_OTHER_SIG_X'),
        RPM(name='sample07', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID fd372689897da07a'),
        RPM(name='sample08', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_OTHER_SIG_X'),
        RPM(name='sample09', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 45689c882fa658e0')]

    current_actor_context.feed(InstalledRPM(items=installed_rpm))
    current_actor_context.run()
    assert current_actor_context.consume(InstalledRedHatSignedRPM)
    assert len(current_actor_context.consume(InstalledRedHatSignedRPM)[0].items) == 5
    assert current_actor_context.consume(InstalledUnsignedRPM)
    assert len(current_actor_context.consume(InstalledUnsignedRPM)[0].items) == 4


def test_gpg_pubkey_pkg(current_actor_context):
    installed_rpm = [
        RPM(name='gpg-pubkey', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID aa17105e03152d37'),
        RPM(name='gpg-pubkey', version='0.1', release='1.sm01', epoch='1', packager='Tester', arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 9ea903b1361e896b'),
    ]

    current_actor_context.feed(InstalledRPM(items=installed_rpm))
    current_actor_context.run()
    assert current_actor_context.consume(InstalledRedHatSignedRPM)
    assert len(current_actor_context.consume(InstalledRedHatSignedRPM)[0].items) == 1
    assert current_actor_context.consume(InstalledUnsignedRPM)
    assert len(current_actor_context.consume(InstalledUnsignedRPM)[0].items) == 1


def test_create_lookup():
    # NOTE(ivasilev) Ideally should be tested separately from the actor, but since library
    # testing functionality is not yet implemented in leapp-repository the tests will reside here.
    model = MockModel()
    # plain non-empty list
    model.list_field.extend([-42])
    with mock.patch('leapp.libraries.stdlib.api.consume', return_value=(model,)):
        # monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_message_mocked)
        lookup = rpms.create_lookup(MockModel, 'list_field', 'real')
        assert {42, -42} == lookup
    # empty list
    model.list_field = []
    with mock.patch('leapp.libraries.stdlib.api.consume', return_value=(model,)):
        lookup = rpms.create_lookup(MockModel, 'list_field', 'real')
        assert {} == lookup
    # nullable list without default
    assert model.list_field_nullable is None
    with mock.patch('leapp.libraries.stdlib.api.consume', return_value=(model,)):
        lookup = rpms.create_lookup(MockModel, 'list_field_nullable', 'real')
        assert {} == lookup
    # improper usage: lookup from non iterable field
    with mock.patch('leapp.libraries.stdlib.api.consume', return_value=(model,)):
        lookup = rpms.create_lookup(MockModel, 'int_field', 'real')
        assert {} == lookup
    # improper usage: lookup from iterable but bad attribute
    with mock.patch('leapp.libraries.stdlib.api.consume', return_value=(model,)):
        lookup = rpms.create_lookup(MockModel, 'list_field', 'nosuchattr')
        assert {} == lookup


def test_has_package(current_actor_context):
    installed_rpm = [
        RPM(name='sample01', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
        RPM(name='sample02', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_OTHER_SIG_X'),
    ]

    current_actor_context.feed(InstalledRPM(items=installed_rpm))
    current_actor_context.run()
    assert rpms.has_package(InstalledRedHatSignedRPM, 'sample01', context=current_actor_context)
    assert not rpms.has_package(InstalledRedHatSignedRPM, 'nosuchpackage', context=current_actor_context)
    assert rpms.has_package(InstalledUnsignedRPM, 'sample02', context=current_actor_context)
    assert not rpms.has_package(InstalledUnsignedRPM, 'nosuchpackage', context=current_actor_context)



from leapp.actors import Actor
from leapp.libraries.actor import systemfacts
from leapp.models import SysctlVariablesFacts, ActiveKernelModulesFacts, UsersFacts, GroupsFacts, RepositoriesFacts, \
    SELinuxFacts, FirewallsFacts
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class SystemFactsActor(Actor):
    """
    Provides data about many facts from system.

    After collecting data from multiple tools, messages with relevant data will be produced to
    describe:
      - Sysctl variables;
      - Active Linux Kernel Modules;
      - Users;
      - Groups;
      - Package repositories;
      - SELinux status;
      - Firewalls status.
    """

    name = 'system_facts'
    consumes = ()
    produces = (SysctlVariablesFacts,
                ActiveKernelModulesFacts,
                UsersFacts,
                GroupsFacts,
                RepositoriesFacts,
                SELinuxFacts,
                FirewallsFacts)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        self.produce(systemfacts.get_sysctls_status())
        self.produce(systemfacts.get_active_kernel_modules_status(self.log))
        self.produce(systemfacts.get_system_users_status())
        self.produce(systemfacts.get_system_groups_status())
        self.produce(systemfacts.get_repositories_status())
        self.produce(systemfacts.get_selinux_status())
        self.produce(systemfacts.get_firewalls_status())



from leapp.libraries.common.reporting import report_with_links
from leapp.libraries.common.tcpwrappersutils import config_applies_to_daemon


def check_config_supported(tcpwrap_facts, vsftpd_facts):
    bad_configs = [config.path for config in vsftpd_facts.configs if config.tcp_wrappers]
    if bad_configs and config_applies_to_daemon(tcpwrap_facts, 'vsftpd'):
        list_separator_fmt = '\n    - '
        report_with_links(title='Unsupported vsftpd configuration',
                          summary=('tcp_wrappers support has been removed in RHEL-8. '
                                   'Some configuration files set the tcp_wrappers option to true and '
                                   'there is some vsftpd-related configuration in /etc/hosts.deny '
                                   'or /etc/hosts.allow. Please migrate it manually. '
                                   'The list of problematic configuration files:{}{}'
                                   ).format(list_separator_fmt,
                                            list_separator_fmt.join(bad_configs)),
                          links=[{'title': 'Replacing TCP Wrappers in RHEL 8',
                                  'href': 'https://access.redhat.com/solutions/3906701'}],
                          severity='high',
                          flags=['inhibitor'])



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import CalledProcessError, api, run
from leapp.models import BootContent


def remove_boot_entry():
    # we need to make sure /boot is mounted before trying to remove the boot entry
    try:
        run([
            '/bin/mount', '/boot'
        ])
    except CalledProcessError:
        # /boot has been most likely already mounted
        pass
    kernel_filepath = get_upgrade_kernel_filepath()
    run([
        '/usr/sbin/grubby',
        '--remove-kernel={0}'.format(kernel_filepath)
    ])
    # TODO: Move calling `mount -a` to a separate actor as it is not really related to removing the upgrade boot entry.
    #       It's worth to call it after removing the boot entry to avoid boot loop in case mounting fails.
    run([
        '/bin/mount', '-a'
    ])


def get_upgrade_kernel_filepath():
    boot_content_msgs = api.consume(BootContent)
    boot_content = next(boot_content_msgs, None)
    if list(boot_content_msgs):
        api.current_logger().warning('Unexpectedly received more than one BootContent message.')
    if not boot_content:
        raise StopActorExecutionError('Could not create a GRUB boot entry for the upgrade initramfs.',
                                      details={'details': 'Did not receive a message about the leapp-provided'
                                                          ' kernel and initramfs.'})
    return boot_content.kernel_path



from leapp.models import Model, fields
from leapp.topics import BootPrepTopic


class UpgradeDracutModule(Model):
    """
    This model is used to influence the leapp upgrade initram disk generation by allowing to
    include dracut modules specified by this message.
    """
    topic = BootPrepTopic

    name = fields.String()
    """ Name of the dracut module that should be added (--add option of dracut) """

    module_path = fields.Nullable(fields.String(default=None))
    """
    module_path specifies dracut modules that are to be copied
    If the path is not set, the given name will just be activated.
    """



import os

from leapp.snactor.fixture import current_actor_context


def test_create_resume_service(current_actor_context):

    current_actor_context.run()

    service_name = 'leapp_resume.service'
    service_path = '/etc/systemd/system/{}'.format(service_name)
    symlink_path = '/etc/systemd/system/default.target.wants/{}'.format(service_name)

    try:
        assert os.path.isfile(service_path)
        assert os.path.isfile(symlink_path)
    finally:
        os.unlink(service_path)
        os.unlink(symlink_path)



import os
import errno

from leapp.actors import Actor
from leapp.libraries.common.reporting import report_generic
from leapp.libraries.stdlib import run
from leapp.reporting import Report
from leapp.tags import FirstBootPhaseTag, IPUWorkflowTag


class RemoveSystemdResumeService(Actor):
    """
    Remove systemd service to launch Leapp.

    After system was rebooted and process resumed, this service is not necessary anymore.
    """

    name = 'remove_systemd_resume_service'
    consumes = ()
    produces = (Report,)
    tags = (FirstBootPhaseTag, IPUWorkflowTag)

    def process(self):
        service_name = 'leapp_resume.service'
        if os.path.isfile('/etc/systemd/system/{}'.format(service_name)):
            run(['systemctl', 'disable', service_name])
            try:
                os.unlink('/etc/systemd/system/{}'.format(service_name))
                os.unlink('/etc/systemd/system/default.target.wants/{}'.format(service_name))
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise

        report_generic(
            title='"{}" service deleted'.format(service_name),
            summary='"{}" was taking care of resuming upgrade process '
                    'after the first reboot.'.format(service_name),
            severity='low'
        )



import os.path
from collections import namedtuple

import pytest

from leapp.libraries import stdlib
from leapp.libraries.actor import forcedefaultboot
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelVersion

Expected = namedtuple(
    'Expected', (
     'grubby_setdefault'
    )
)

Case = namedtuple(
    'Case',
    ('initrd_exists',
     'kernel_exists',
     'entry_default',
     'entry_exists',
     'message_available'
     )
)

TARGET_KERNEL_VERSION = '1.2.3.4.el8.x86_64'
TARGET_KERNEL_PATH = '/boot/vmlinuz-{}'.format(TARGET_KERNEL_VERSION)
TARGET_INITRD_PATH = '/boot/initramfs-{}.img'.format(TARGET_KERNEL_VERSION)

OLD_KERNEL_VERSION = '0.1.2.3.el7.x86_64'
OLD_KERNEL_PATH = '/boot/vmlinuz-{}'.format(OLD_KERNEL_VERSION)

CASES = (
    (Case(initrd_exists=True, kernel_exists=True, entry_default=True, entry_exists=True, message_available=True),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=False, kernel_exists=True, entry_default=False, entry_exists=True, message_available=True),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=True, kernel_exists=False, entry_default=False, entry_exists=True, message_available=True),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=False, kernel_exists=False, entry_default=False, entry_exists=True, message_available=True),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=True, kernel_exists=True, entry_default=False, entry_exists=True, message_available=False),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=True, kernel_exists=True, entry_default=False, entry_exists=False, message_available=False),
     Expected(grubby_setdefault=False)),
    (Case(initrd_exists=True, kernel_exists=True, entry_default=False, entry_exists=True, message_available=True),
     Expected(grubby_setdefault=True))
)

_GRUBBY_INFO_TEMPLATE = '''index={entry_index}
kernel=/boot/vmlinuz-{kernel_version}
args="ro rd.lvm.lv=testing/root rd.lvm.lv=testing/swap rhgb quiet LANG=en_US.UTF-8"
root=/dev/mapper/testing-root
initrd=/boot/initramfs-{kernel_version}.img
'''


class MockedRun(object):
    def __init__(self, case):
        self.case = case
        self.called_setdefault = False

    def __call__(self, cmd, *args, **kwargs):
        if cmd and cmd[0] == 'grubby':
            target = getattr(self, 'grubby_{}'.format(cmd[1].strip('--').replace('-', '_')), None)
            assert target and 'Unsupport grubby command called'
            return target(cmd)  # pylint: disable=not-callable
        return None

    def grubby_info(self, cmd):
        assert len(cmd) == 3
        if not self.case.entry_exists:
            raise stdlib.CalledProcessError('A leapp command failed', cmd, {})
        else:
            return {
                'stdout': _GRUBBY_INFO_TEMPLATE.format(
                    entry_index=0 if self.case.entry_default else 1,
                    kernel_version=TARGET_KERNEL_VERSION)
                }

    def grubby_default_kernel(self, cmd):
        assert len(cmd) == 2
        if self.case.entry_default:
            return {'stdout': '{}\n'.format(TARGET_KERNEL_PATH)}
        return {'stdout': '{}\n'.format(OLD_KERNEL_PATH)}

    def grubby_set_default(self, cmd):
        assert len(cmd) == 3
        assert cmd[2] == TARGET_KERNEL_PATH
        self.called_setdefault = True


def mocked_consume(case):
    def impl(*args):
        if case.message_available:
            return iter((InstalledTargetKernelVersion(version=TARGET_KERNEL_VERSION),))
        return iter(())
    return impl


def mocked_exists(case, orig_path_exists):
    def impl(path):
        if path == TARGET_KERNEL_PATH:
            return case.kernel_exists
        if path == TARGET_INITRD_PATH:
            return case.initrd_exists
        return orig_path_exists(path)
    return impl


@pytest.mark.parametrize('case_result', CASES)
def test_force_default_boot_target_scenario(case_result, monkeypatch):
    case, result = case_result
    mocked_run = MockedRun(case)
    monkeypatch.setattr(api, 'consume', mocked_consume(case))
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(os.path, 'exists', mocked_exists(case, os.path.exists))
    forcedefaultboot.process()
    assert result.grubby_setdefault == mocked_run.called_setdefault



# plugin inspired by "system_upgrade.py" from rpm-software-management
from __future__ import print_function

import json

import dnf
import dnf.cli

CMDS = ['download', 'upgrade', 'check']


class DoNotDownload(Exception):
    pass


def _do_not_download_packages(packages, progress=None, total=None):
    raise DoNotDownload()


class RhelUpgradeCommand(dnf.cli.Command):
    aliases = ('rhel-upgrade',)
    summary = ("Plugin for upgrading to the next RHEL major release")

    def __init__(self, cli):
        super(RhelUpgradeCommand, self).__init__(cli)
        self.plugin_data = {}
        self.pkgs_notfound = []

    @staticmethod
    def set_argparser(parser):
        parser.add_argument('tid', nargs=1, choices=CMDS,
                            metavar="[%s]" % "|".join(CMDS))
        parser.add_argument('filename')

    def pre_configure(self):
        with open(self.opts.filename) as fo:
            self.plugin_data = json.load(fo)
        # There is an issue that ignores releasever value if it is set at configure
        self.base.conf.releasever = self.plugin_data['dnf_conf']['releasever']

    def configure(self):
        self.cli.demands.root_user = True
        self.cli.demands.resolving = self.opts.tid[0] != 'check'
        self.cli.demands.available_repos = True
        self.cli.demands.sack_activation = True
        self.cli.demands.cacheonly = self.opts.tid[0] == 'upgrade'
        self.cli.demands.allow_erasing = self.plugin_data['dnf_conf']['allow_erasing']
        self.base.conf.protected_packages = []
        self.base.conf.best = self.plugin_data['dnf_conf']['best']
        self.base.conf.assumeyes = True
        self.base.conf.gpgcheck = self.plugin_data['dnf_conf']['gpgcheck']
        self.base.conf.debug_solver = self.plugin_data['dnf_conf']['debugsolver']
        self.base.conf.module_platform_id = self.plugin_data['dnf_conf']['platform_id']
        installroot = self.plugin_data['dnf_conf'].get('installroot')
        if installroot:
            self.base.conf.installroot = installroot
        if self.plugin_data['dnf_conf']['test_flag'] and self.opts.tid[0] == 'download':
            self.base.conf.tsflags.append("test")

        enabled_repos = self.plugin_data['dnf_conf']['enable_repos']
        self.base.repos.all().disable()
        for repo in self.base.repos.all():
            if repo.id in enabled_repos:
                repo.skip_if_unavailable = False
                if not self.base.conf.gpgcheck:
                    repo.gpgcheck = False
                repo.enable()

    def run(self):
        self.base.add_remote_rpms(self.plugin_data['pkgs_info']['local_rpms'])

        for pkg_spec in self.plugin_data['pkgs_info']['to_remove']:
            try:
                self.base.remove(pkg_spec)
            except dnf.exceptions.MarkingError:
                self.pkgs_notfound.append(pkg_spec)

        for pkg_spec in self.plugin_data['pkgs_info']['to_install']:
            try:
                self.base.install(pkg_spec)
            except dnf.exceptions.MarkingError:
                self.pkgs_notfound.append(pkg_spec)

        q = self.base.sack.query().installed()
        for pkg in q:
            if pkg.name not in (self.plugin_data['pkgs_info']['to_install']
                                + self.plugin_data['pkgs_info']['to_remove']):
                self.base.upgrade(pkg.name)

        self.base.distro_sync()

        if self.opts.tid[0] == 'check':
            self.base.resolve(allow_erasing=self.cli.demands.allow_erasing)

            # We are doing this to avoid downloading the packages in the check phase
            self.base.download_packages = _do_not_download_packages
            try:
                self.base.do_transaction()
            except DoNotDownload:
                print('Check completed.')


class RhelUpgradePlugin(dnf.Plugin):
    name = 'rhel-upgrade'

    def __init__(self, base, cli):
        super(RhelUpgradePlugin, self).__init__(base, cli)
        if cli:
            cli.register_command(RhelUpgradeCommand)



from leapp.actors import Actor
from leapp.models import InstalledRedHatSignedRPM
from leapp.libraries.common.reporting import report_with_remediation
from leapp.libraries.common.rpms import has_package
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckAcpid(Actor):
    """
    Check if acpid is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkacpid'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(InstalledRedHatSignedRPM, 'acpid'):
            report_with_remediation(
                title='Acpid incompatible changes in the next major version',
                summary='The option -d (debug) no longer implies -f (foreground).',
                remediation='You must now use both options (\'-df\') for the same behavior. Please update '
                            'your scripts to be compatible with the changes.',
                severity='low')



from leapp.libraries.actor import library
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import report_generic_mocked
from leapp.libraries.stdlib import api
from leapp.models import PartitionEntry, FstabEntry, MountEntry, LsblkEntry, PvsEntry, VgsEntry, \
    LvdisplayEntry, SystemdMountEntry


def test_get_partitions_info(monkeypatch):
    def is_file_readable_mocked(path):
        return False

    expected = [
        PartitionEntry(major='252', minor='0', blocks='41943040', name='vda'),
        PartitionEntry(major='252', minor='1', blocks='1048576', name='vda1'),
        PartitionEntry(major='252', minor='2', blocks='40893440', name='vda2'),
        PartitionEntry(major='253', minor='0', blocks='39837696', name='dm-0'),
        PartitionEntry(major='253', minor='1', blocks='1048576', name='dm-1')]
    assert expected == library._get_partitions_info('tests/files/partitions')

    monkeypatch.setattr(library, '_is_file_readable', is_file_readable_mocked)
    assert [] == library._get_partitions_info('unreadable_file')


def test_get_fstab_info(monkeypatch):
    expected = [
        FstabEntry(
            fs_spec='/dev/mapper/rhel_ibm--p8--kvm--03--guest--02-root',
            fs_file='/',
            fs_vfstype='xfs',
            fs_mntops='defaults',
            fs_freq='0',
            fs_passno='0'),
        FstabEntry(
            fs_spec='UUID=0a5215ef-1fb4-4b1b-8860-be4baa9e624c',
            fs_file='/boot',
            fs_vfstype='xfs',
            fs_mntops='defaults',
            fs_freq='0',
            fs_passno='1'),
        FstabEntry(
            fs_spec='UUID=acf9f525-3691-429f-96d7-3f8530227062',
            fs_file='/var',
            fs_vfstype='xfs',
            fs_mntops='defaults',
            fs_freq='0',
            fs_passno='0'),
        FstabEntry(
            fs_spec='UUID=d74186c9-21d5-4549-ae26-91ca9ed36f56',
            fs_file='/tmp',
            fs_vfstype='ext4',
            fs_mntops='defaults,nodev,nosuid,noexec',
            fs_freq='1',
            fs_passno='0'),
        FstabEntry(
            fs_spec='/dev/mapper/rhel_ibm--p8--kvm--03--guest--02-swap',
            fs_file='swap',
            fs_vfstype='swap',
            fs_mntops='defaults',
            fs_freq='0',
            fs_passno='0')]
    assert expected == library._get_fstab_info('tests/files/fstab')
    monkeypatch.setattr(library, '_is_file_readable', lambda(_): False)
    assert [] == library._get_fstab_info('unreadable_file')


def test_invalid_fstab_info(monkeypatch):
    class logger_mocked(object):
        def __init__(self):
            self.errmsg = None

        def error(self, msg):
            self.errmsg = msg

        def __call__(self):
            return self

    monkeypatch.setattr(reporting, "report_with_remediation", report_generic_mocked())
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    library._get_fstab_info('tests/files/invalid_fstab')
    assert reporting.report_with_remediation.called == 1
    assert reporting.report_with_remediation.report_fields['severity'] == 'high'
    assert 'Problems with parsing data in /etc/fstab' in reporting.report_with_remediation.report_fields['title']
    assert 'inhibitor' in reporting.report_with_remediation.report_fields['flags']
    assert "The fstab configuration file seems to be invalid" in api.current_logger.errmsg


def test_get_mount_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['sysfs', 'on', '/sys', 'type', 'sysfs', '(rw,nosuid,nodev,noexec,relatime,seclabel)'],
            ['proc', 'on', '/proc', 'type', 'proc', '(rw,nosuid,nodev,noexec,relatime)'],
            ['tmpfs', 'on', '/dev/shm', 'type', 'tmpfs', '(rw,nosuid,nodev,seclabel)'],
            ['tmpfs', 'on', '/run', 'type', 'tmpfs', '(rw,nosuid,nodev,seclabel,mode=755)']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        MountEntry(
            name='sysfs',
            mount='/sys',
            tp='sysfs',
            options='(rw,nosuid,nodev,noexec,relatime,seclabel)'),
        MountEntry(
            name='proc',
            mount='/proc',
            tp='proc',
            options='(rw,nosuid,nodev,noexec,relatime)'),
        MountEntry(
            name='tmpfs',
            mount='/dev/shm',
            tp='tmpfs',
            options='(rw,nosuid,nodev,seclabel)'),
        MountEntry(
            name='tmpfs',
            mount='/run',
            tp='tmpfs',
            options='(rw,nosuid,nodev,seclabel,mode=755)')]
    assert expected == library._get_mount_info()


def test_get_lsblk_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['vda', '252:0', '0', '40G', '0', 'disk', ''],
            ['vda1', '252:1', '0', '1G', '0', 'part', '/boot'],
            ['vda2', '252:2', '0', '39G', '0', 'part', ''],
            ['rhel_ibm--p8--kvm--03--guest--02-root', '253:0', '0', '38G', '0', 'lvm', '/'],
            ['rhel_ibm--p8--kvm--03--guest--02-swap', '253:1', '0', '1G', '0', 'lvm', '[SWAP]']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        LsblkEntry(
            name='vda',
            maj_min='252:0',
            rm='0',
            size='40G',
            ro='0',
            tp='disk',
            mountpoint=''),
        LsblkEntry(
            name='vda1',
            maj_min='252:1',
            rm='0',
            size='1G',
            ro='0',
            tp='part',
            mountpoint='/boot'),
        LsblkEntry(
            name='vda2',
            maj_min='252:2',
            rm='0',
            size='39G',
            ro='0',
            tp='part',
            mountpoint=''),
        LsblkEntry(
            name='rhel_ibm--p8--kvm--03--guest--02-root',
            maj_min='253:0',
            rm='0',
            size='38G',
            ro='0',
            tp='lvm',
            mountpoint='/'),
        LsblkEntry(
            name='rhel_ibm--p8--kvm--03--guest--02-swap',
            maj_min='253:1',
            rm='0',
            size='1G',
            ro='0',
            tp='lvm',
            mountpoint='[SWAP]')]
    assert expected == library._get_lsblk_info()


def test_get_pvs_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['/dev/vda2', 'rhel_ibm-p8-kvm-03-guest-02', 'lvm2', 'a--', '<39.00g', '4.00m']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        PvsEntry(
            pv='/dev/vda2',
            vg='rhel_ibm-p8-kvm-03-guest-02',
            fmt='lvm2',
            attr='a--',
            psize='<39.00g',
            pfree='4.00m')]
    assert expected == library._get_pvs_info()


def test_get_vgs_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['rhel_ibm-p8-kvm-03-guest-02', '1', '2', '0', 'wz--n-', '<39.00g', '4.00m']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        VgsEntry(
            vg='rhel_ibm-p8-kvm-03-guest-02',
            pv='1',
            lv='2',
            sn='0',
            attr='wz--n-',
            vsize='<39.00g',
            vfree='4.00m')]
    assert expected == library._get_vgs_info()


def test_get_lvdisplay_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['root', 'rhel_ibm-p8-kvm-03-guest-02', '-wi-ao----', '37.99g', '', '', '', '', '', '', '', ''],
            ['swap', 'rhel_ibm-p8-kvm-03-guest-02', '-wi-ao----', '1.00g', '', '', '', '', '', '', '', '']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        LvdisplayEntry(
            lv='root',
            vg='rhel_ibm-p8-kvm-03-guest-02',
            attr='-wi-ao----',
            lsize='37.99g',
            pool='',
            origin='',
            data='',
            meta='',
            move='',
            log='',
            cpy_sync='',
            convert=''),
        LvdisplayEntry(
            lv='swap',
            vg='rhel_ibm-p8-kvm-03-guest-02',
            attr='-wi-ao----',
            lsize='1.00g',
            pool='',
            origin='',
            data='',
            meta='',
            move='',
            log='',
            cpy_sync='',
            convert='')]
    assert expected == library._get_lvdisplay_info()


def test_get_systemd_mount_info(monkeypatch):
    def get_cmd_output_mocked(cmd, delim, expected_len):
        return [
            ['/dev/dm-1',
             'n/a',
             'n/a',
             'n/a',
             'ext4',
             'n/a',
             'bec30ca5-5403-4c23-ae6e-cb2a911bc076'],
            ['/dev/dm-3',
             'n/a',
             'n/a',
             'n/a',
             'ext4',
             'n/a',
             'd6eaf17d-e2a9-4e8d-bb54-a89c18923ea2'],
            ['/dev/sda1',
             'pci-0000:00:17.0-ata-2',
             'LITEON_LCH-256V2S',
             '0x5002303100d82b06',
             'ext4',
             'n/a',
             'c3890bf3-9273-4877-ad1f-68144e1eb858']]

    monkeypatch.setattr(library, '_get_cmd_output', get_cmd_output_mocked)
    expected = [
        SystemdMountEntry(
            node='/dev/dm-1',
            path='n/a',
            model='n/a',
            wwn='n/a',
            fs_type='ext4',
            label='n/a',
            uuid='bec30ca5-5403-4c23-ae6e-cb2a911bc076'),
        SystemdMountEntry(
            node='/dev/dm-3',
            path='n/a',
            model='n/a',
            wwn='n/a',
            fs_type='ext4',
            label='n/a',
            uuid='d6eaf17d-e2a9-4e8d-bb54-a89c18923ea2'),
        SystemdMountEntry(
            node='/dev/sda1',
            path='pci-0000:00:17.0-ata-2',
            model='LITEON_LCH-256V2S',
            wwn='0x5002303100d82b06',
            fs_type='ext4',
            label='n/a',
            uuid='c3890bf3-9273-4877-ad1f-68144e1eb858')]
    assert expected == library._get_systemd_mount_info()



from leapp.snactor.fixture import current_actor_context
from leapp.models import StorageInfo, SystemdMountEntry, FstabEntry, MountEntry
from leapp.reporting import Report


def test_actor_with_systemdmount_entry(current_actor_context):
    with_systemdmount_entry = [SystemdMountEntry(node="nfs", path="n/a", model="n/a",
                                                 wwn="n/a", fs_type="nfs", label="n/a",
                                                 uuid="n/a")]
    current_actor_context.feed(StorageInfo(systemdmount=with_systemdmount_entry))
    current_actor_context.run()
    assert 'inhibitor' in current_actor_context.consume(Report)[0].flags


def test_actor_without_systemdmount_entry(current_actor_context):
    without_systemdmount_entry = [SystemdMountEntry(node="/dev/sda1",
                                                    path="pci-0000:00:17.0-ata-2",
                                                    model="TOSHIBA_THNSNJ512GDNU_A",
                                                    wwn="0x500080d9108e8753",
                                                    fs_type="ext4", label="n/a",
                                                    uuid="5675d309-eff7-4eb1-9c27-58bc5880ec72")]
    current_actor_context.feed(StorageInfo(systemdmount=without_systemdmount_entry))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_with_fstab_entry(current_actor_context):
    with_fstab_entry = [FstabEntry(fs_spec="lithium:/mnt/data", fs_file="/mnt/data",
                                   fs_vfstype="nfs",
                                   fs_mntops="noauto,noatime,rsize=32768,wsize=32768",
                                   fs_freq="0", fs_passno="0")]
    current_actor_context.feed(StorageInfo(fstab=with_fstab_entry))
    current_actor_context.run()
    assert 'inhibitor' in current_actor_context.consume(Report)[0].flags


def test_actor_without_fstab_entry(current_actor_context):
    without_fstab_entry = [FstabEntry(fs_spec="/dev/mapper/fedora-home", fs_file="/home",
                                      fs_vfstype="ext4",
                                      fs_mntops="defaults,x-systemd.device-timeout=0",
                                      fs_freq="1", fs_passno="2")]
    current_actor_context.feed(StorageInfo(fstab=without_fstab_entry))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_with_mount_share(current_actor_context):
    with_mount_share = [MountEntry(name="nfs", mount="/mnt/data", tp="nfs",
                                   options="rw,nosuid,nodev,relatime,user_id=1000,group_id=1000")]
    current_actor_context.feed(StorageInfo(mount=with_mount_share))
    current_actor_context.run()
    assert 'inhibitor' in current_actor_context.consume(Report)[0].flags


def test_actor_without_mount_share(current_actor_context):
    without_mount_share = [MountEntry(name="tmpfs", mount="/run/snapd/ns", tp="tmpfs",
                                      options="rw,nosuid,nodev,seclabel,mode=755")]
    current_actor_context.feed(StorageInfo(mount=without_mount_share))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



from leapp.models import Model, fields
from leapp.topics import SystemFactsTopic


class WhitelistedKernelModules(Model):
    """
    whitelisted_modules: list of module names that are considered whitelisted
    and are not going to inhibit the upgrade
    """
    topic = SystemFactsTopic
    whitelisted_modules = fields.List(fields.String(), default=[])



from six.moves.configparser import ConfigParser, ParsingError

from leapp.actors import Actor
from leapp.libraries.stdlib import CalledProcessError, run
from leapp.models import NetworkManagerConfig
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class NetworkManagerReadConfig(Actor):
    """
    Provides data about NetworkManager configuration.

    After collecting data from NetworkManager tool, a message with relevant data will be produced.
    """

    name = 'network_manager_read_config'
    consumes = ()
    produces = (NetworkManagerConfig,)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        nm_config = NetworkManagerConfig()
        try:
            # Use 'NM --print-config' to read the configurationo so
            # that the main configuration file and other files in
            # various directories get merged in the right way.
            r = run(['NetworkManager', '--print-config'], split=False)['stdout']
        except (OSError, CalledProcessError) as e:
            self.log.warning('Error reading NetworkManager configuration: {}'.format(e))
            return

        parser = ConfigParser()

        try:
            if hasattr(parser, 'read_string'):
                # Python 3
                parser.read_string(r)
            else:
                # Python 2
                from cStringIO import StringIO
                parser.readfp(StringIO(r))
        except ParsingError as e:
            self.log.warning('Error parsing NetworkManager configuration: {}'.format(e))
            return

        if parser.has_option('main', 'dhcp'):
            nm_config.dhcp = parser.get("main", "dhcp")

        self.produce(nm_config)



from leapp.libraries.actor.systemfacts import anyhasprefix, anyendswith, aslist
from leapp.snactor.fixture import current_actor_libraries


def test_anyendswith(current_actor_libraries):
    value = 'this_is_a_test'

    assert anyendswith(value, ['a_test'])
    assert anyendswith(value, ['a_test', 'bgerwh', 'g52h4q'])
    assert anyendswith(value, ['est'])
    assert not anyendswith(value, ['asafsaf', 'gbfdshh', '123f', 'gdsgsnb'])
    assert not anyendswith(value, [])


def test_anyhasprefix(current_actor_libraries):
    value = 'this_is_a_test'

    assert anyhasprefix(value, ['this'])
    assert anyhasprefix(value, ['this', 'ddsvssd', 'bsdhn', '125fff'])
    assert anyhasprefix(value, ['this_is'])
    assert not anyhasprefix(value, ['ccbbb', 'xasbnn', 'xavavav', 'bbnkk1'])
    assert not anyhasprefix(value, [])


def test_aslist(current_actor_libraries):

    @aslist
    def local():
        yield True
        yield False
        yield True

    r = local()

    assert isinstance(r, list) and r[0] and r[2] and not r[1]



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_with_remediation
from leapp.libraries.common.rpms import has_package
from leapp.models import InstalledRedHatSignedRPM
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.reporting import Report


class CheckIrssi(Actor):
    """
    Check if irssi is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkirssi'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(InstalledRedHatSignedRPM, 'irssi'):
            report_with_remediation(
                title='Irssi incompatible changes in the next major version',
                summary='Disabled support for the insecure SSLv2 protocol.\n'
                        'Disabled SSLv3 due to the POODLE vulnerability.\n'
                        'Removing networks will now remove all attached servers and channels.\n'
                        'Removed --disable-ipv6 option.\n',
                remediation='Please update your scripts to be compatible with the changes.',
                severity='low')



from leapp.libraries.actor import library
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import report_generic_mocked


def test_uninstalled(monkeypatch):
    for config_default in (False, True):
        monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
        monkeypatch.setattr(library, 'is_config_default', lambda: config_default)

        library.check_chrony(False)

        assert reporting.report_generic.called == 0


def test_installed_defconf(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setattr(library, 'is_config_default', lambda: True)

    library.check_chrony(True)

    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'chrony using default configuration'


def test_installed_nodefconf(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setattr(library, 'is_config_default', lambda: False)

    library.check_chrony(True)

    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'chrony using non-default configuration'



from leapp.models import Model, fields
from leapp.topics import TransactionTopic


class TargetUserSpaceInfo(Model):
    topic = TransactionTopic
    path = fields.String()
    scratch = fields.String()
    mounts = fields.String()



import re

from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api, run


sysconfig_path = '/etc/sysconfig/memcached'


def is_sysconfig_default():
    """Check if the memcached sysconfig file was not modified since installation."""
    try:
        result = run(['rpm', '-V', '--nomtime', 'memcached'], checked=False)
        return sysconfig_path not in result['stdout']
    except OSError as e:
        api.current_logger().warn("rpm verification failed: %s", str(e))
        return True


def is_udp_disabled():
    """Check if UDP port is disabled in the sysconfig file."""
    with open(sysconfig_path) as f:
        for line in f:
            if re.match(r'^\s*OPTIONS=.*-U\s*0[^0-9]', line):
                return True
    return False


def check_memcached(memcached_installed):
    """Report potential issues in memcached configuration."""
    if not memcached_installed:
        api.current_logger().info('memcached package is not installed')
        return

    default_memcached_conf = is_sysconfig_default()
    disabled_udp_port = is_udp_disabled()

    if default_memcached_conf:
        reporting.report_generic(
            title='memcached service is using default configuration',
            summary='memcached in RHEL8 listens on loopback only and has UDP port disabled by default',
            severity='medium')
    elif not disabled_udp_port:
        reporting.report_generic(
            title='memcached has enabled UDP port',
            summary='memcached in RHEL7 has UDP port enabled by default, but it is disabled by default in RHEL8',
            severity='medium')
    else:
        reporting.report_generic(
            title='memcached has already disabled UDP port',
            summary='memcached in RHEL8 has UDP port disabled by default',
            severity='low')



import os.path
import re

from leapp.libraries.actor import modscan
from leapp.libraries.stdlib import api
from leapp.models import UpgradeDracutModule, RequiredUpgradeInitramPackages


def _files_get_folder_path(name):
    path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'files', name)
    assert os.path.isdir(path)
    return path


def test_created_modules(monkeypatch):
    with monkeypatch.context() as context:
        context.setattr(api, 'get_actor_folder_path', _files_get_folder_path)
        path = os.path.abspath(api.get_actor_folder_path('dracut'))
        required_modules = ['sys-upgrade', 'sys-upgrade-redhat']
        for mod in modscan._create_dracut_modules():
            index = required_modules.index(mod.name)

            # Ensures that this actor only includes known modules
            assert index != -1

            required_modules.pop(index)

            # Ensures that the path is valid
            assert mod.module_path
            assert mod.module_path == os.path.abspath(mod.module_path)

            # Ensure it's a directory
            assert os.path.isdir(mod.module_path)

            # Ensure it's located within the actors files path
            assert mod.module_path.startswith(path)

            # Ensure the directory name ends with the module name
            assert os.path.basename(mod.module_path).endswith(mod.name)
            assert not re.sub(r'^(85|90){}$'.format(mod.name), '', os.path.basename(mod.module_path))
        assert not required_modules


def test_required_packages():
    assert set(modscan._REQUIRED_PACKAGES) == set(modscan._create_initram_packages().packages)


def test_process_produces_modules(monkeypatch):
    with monkeypatch.context() as context:
        messages = []
        context.setattr(api, 'produce', lambda *x: messages.extend(x))
        context.setattr(api, 'get_actor_folder_path', _files_get_folder_path)
        modscan.process()
        assert messages
        assert len(messages) == 3
        assert len([msg for msg in messages if isinstance(msg, UpgradeDracutModule)]) == 2
        assert len([msg for msg in messages if isinstance(msg, RequiredUpgradeInitramPackages)]) == 1



import os.path

from leapp.models import RpmTransactionTasks


def load_tasks_file(path, logger):
    # Loads the given file and converts it to a deduplicated list of strings that are stripped
    if os.path.isfile(path):
        try:
            with open(path, 'r') as f:
                return list(
                    {entry.strip() for entry in f.read().split('\n') if entry.strip() and
                        not entry.strip().startswith('#')}
                )
        except IOError as e:
            logger.warn('Failed to open %s to load additional transaction data. Error: %s', path, str(e))
    return []


def load_tasks(base_dir, logger):
    # Loads configuration files to_install, to_keep, and to_remove from the given base directory
    return RpmTransactionTasks(
        to_install=load_tasks_file(os.path.join(base_dir, 'to_install'), logger),
        to_keep=load_tasks_file(os.path.join(base_dir, 'to_keep'), logger),
        to_remove=load_tasks_file(os.path.join(base_dir, 'to_remove'), logger))



from leapp.libraries.actor import library
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import report_generic_mocked


def test_uninstalled(monkeypatch):
    for sysconfig_default in (False, True):
        for udp_disabled in (False, True):
            monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
            monkeypatch.setattr(library, 'is_sysconfig_default', lambda: sysconfig_default)
            monkeypatch.setattr(library, 'is_udp_disabled', lambda: udp_disabled)

            library.check_memcached(False)

            assert reporting.report_generic.called == 0


def test_installed_defconf(monkeypatch):
    for udp_disabled in (False, True):
        monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
        monkeypatch.setattr(library, 'is_sysconfig_default', lambda: True)
        monkeypatch.setattr(library, 'is_udp_disabled', lambda: udp_disabled)

        library.check_memcached(True)

        assert reporting.report_generic.called == 1
        assert reporting.report_generic.report_fields['title'] == 'memcached service is using default configuration'


def test_installed_nodefconf_udp(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setattr(library, 'is_sysconfig_default', lambda: False)
    monkeypatch.setattr(library, 'is_udp_disabled', lambda: False)

    library.check_memcached(True)

    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'memcached has enabled UDP port'


def test_installed_nodefconf_noudp(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    monkeypatch.setattr(library, 'is_sysconfig_default', lambda: False)
    monkeypatch.setattr(library, 'is_udp_disabled', lambda: True)

    library.check_memcached(True)

    assert reporting.report_generic.called == 1
    assert reporting.report_generic.report_fields['title'] == 'memcached has already disabled UDP port'



from leapp.actors import Actor
from leapp.models import FirewalldFacts
from leapp.libraries.actor import private
from leapp.libraries.common.reporting import report_with_remediation
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckFirewalld(Actor):
    """
    Check for certain firewalld configuration that may prevent an upgrade.
    """

    name = 'check_firewalld'
    consumes = (FirewalldFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        unsupported_tables = []
        unsupported_ipset_types = []
        list_separator_fmt = '\n    -'
        for facts in self.consume(FirewalldFacts):
            for table in facts.ebtablesTablesInUse:
                if not private.isEbtablesTableSupported(table):
                    unsupported_tables.append(table)
            for ipset_type in facts.ipsetTypesInUse:
                if not private.isIpsetTypeSupportedByNftables(ipset_type):
                    unsupported_ipset_types.append(ipset_type)

        if unsupported_tables:
            format_tuple = (
                list_separator_fmt,
                list_separator_fmt.join(list(set(unsupported_tables))),)
            report_with_remediation(
                title='Firewalld is using an unsupported ebtables table.',
                summary='ebtables in RHEL-8 does not support these tables:{}{}'.format(*format_tuple),
                remediation='Remove firewalld direct rules that use these ebtables tables:{}{}'.format(*format_tuple),
                severity='high',
                flags=['inhibitor'])
        if unsupported_ipset_types:
            format_tuple = (
                list_separator_fmt,
                list_separator_fmt.join(list(set(unsupported_ipset_types))),)
            report_with_remediation(
                title='Firewalld is using an unsupported ipset type.',
                summary='These ipset types are not supported by firewalld\'s nftables backend:{}{}'.format(
                    *format_tuple),
                remediation='Remove ipsets of these types from firewalld:{}{}'.format(*format_tuple),
                severity='high',
                flags=['inhibitor'])



import os

from leapp.actors import Actor
from leapp.models import RpmTransactionTasks
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class TransactionWorkarounds(Actor):
    """
    Provides additional RPM transaction tasks based on bundled RPM packages.

    After collecting bundled RPM packages, a message with relevant data will be produced.
    """

    name = 'transaction_workarounds'
    consumes = ()
    produces = (RpmTransactionTasks,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        location = self.get_folder_path('bundled-rpms')
        local_rpms = []
        for name in os.listdir(location):
            if name.endswith('.rpm'):
                # It is important to put here the realpath to the files here, because
                # symlinks cannot be resolved properly inside of the target userspace since they use the /installroot
                # mount target
                local_rpms.append(os.path.realpath(os.path.join(location, name)))
        if local_rpms:
            self.produce(RpmTransactionTasks(local_rpms=local_rpms))



from leapp.actors import Actor
from leapp.models import WhitelistedKernelModules
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class WhitelistKernelModules(Actor):
    """
    Actor produces message that contains list of kernel modules that are
    present on the RHEL7 system but missing on the RHEL8. However, these
    specific modules are considered whitelisted and CheckKernelModules
    actor will take this into account and will not inhibit upgrade
    in case any of the whitelisted modules are loaded.
    """
    name = "whitelist_kernel_modules"
    consumes = ()
    produces = (WhitelistedKernelModules,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        # These modules are present on the clean RHEL7.6 installation
        # in virtual machine via vagrant. They are whitelisted because
        # we do not want to inhibit upgrade from clean installation.
        whitelisted_modules = [
            'ablk_helper', 'crct10dif_common', 'cryptd', 'floppy',
            'gf128mul', 'glue_helper', 'iosf_mbi', 'pata_acpi', 'virtio',
            'virtio_pci', 'virtio_ring'
        ]
        self.produce(WhitelistedKernelModules(whitelisted_modules=whitelisted_modules))



from leapp.actors import Actor
from leapp.libraries.actor.library import semantics_changes
from leapp.models import Report, OpenSshConfig
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api
from leapp.libraries.common.reporting import report_with_remediation


class OpenSshPermitRootLoginCheck(Actor):
    """
    OpenSSH no longer allows root logins with password.

    Check the values of PermitRootLogin in OpenSSH server configuration file
    and warn about potential issues after update.
    """
    name = 'openssh_permit_root_login'
    consumes = (OpenSshConfig, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag, )

    def process(self):
        openssh_messages = self.consume(OpenSshConfig)
        config = next(openssh_messages)
        if list(openssh_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')

        if not config.permit_root_login:
            # TODO find out whether the file was modified and will be
            # replaced by the update. If so, this message is bogus
            report_with_remediation(
                title='Possible problems with remote login using root account',
                summary='OpenSSH configuration file does not explicitly state '
                        'the option PermitRootLogin in sshd_config file, '
                        'which will default in RHEL8 to "prohibit-password".',
                remediation='If you depend on remote root logins using '
                            'passwords, consider setting up a different '
                            'user for remote administration or adding '
                            '"PermitRootLogin yes" to sshd_config.',
                severity='high',
                flags=['inhibitor'])

        # Check if there is at least one PermitRootLogin other than "no"
        # in match blocks (other than Match All).
        # This usually means some more complicated setup depending on the
        # default value being globally "yes" and being overwritten by this
        # match block
        if semantics_changes(config):
            report_with_remediation(
                title='OpenSSH configured to allow root login',
                summary='OpenSSH is configured to deny root logins in match '
                        'blocks, but not explicitly enabled in global or '
                        '"Match all" context. This update changes the '
                        'default to disable root logins using paswords '
                        'so your server migth get inaccessible.',
                remediation='Consider using different user for administrative '
                            'logins or make sure your configration file '
                            'contains the line "PermitRootLogin yes" '
                            'in global context if desired.',
                severity='high',
                flags=['inhibitor'])



import platform

from leapp.actors import Actor
from leapp.libraries.actor import library
from leapp.models import IPUConfig, Version
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag


class IPUWorkflowConfig(Actor):
    """
    IPU workflow config actor
    """

    name = 'ipu_workflow_config'
    consumes = ()
    produces = (IPUConfig, Report)
    tags = (IPUWorkflowTag,)

    def process(self):
        target_version = '8.0'
        os_release = library.get_os_release('/etc/os-release')
        self.produce(IPUConfig(
            leapp_env_vars=library.get_env_vars(),
            os_release=os_release,
            architecture=platform.machine(),
            version=Version(source=os_release.version_id, target=target_version)
        ))



from leapp.actors import Actor
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.models import SELinuxFacts, SelinuxPermissiveDecision, SelinuxRelabelDecision
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_generic


class CheckSelinux(Actor):
    """
    Check SELinux status and produce decision messages for further action.

    Based on SELinux status produces decision messages to relabeling and changing status if
    necessary
    """

    name = 'check_se_linux'
    consumes = (SELinuxFacts,)
    produces = (Report, SelinuxPermissiveDecision, SelinuxRelabelDecision)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):

        fact = next(self.consume(SELinuxFacts), None)
        if not fact:
            return

        enabled = fact.enabled
        conf_status = fact.static_mode

        if conf_status == 'disabled':
            if enabled:
                report_generic(
                    title='SElinux disabled in configuration file but currently enabled',
                    summary='This message is to inform user about non-standard SElinux configuration.',
                    severity='low')
            report_generic(
                title='SElinux disabled',
                summary='SElinux disabled, continuing...',
                severity='low')
            return

        if conf_status in ('enforcing', 'permissive'):
            self.produce(SelinuxRelabelDecision(
                set_relabel=True))
            report_generic(
                title='Schedule SElinux relabeling',
                summary='Schedule SElinux relabeling as the status was permissive/enforcing.',
                severity='low')

        if conf_status == 'enforcing':
            self.produce(SelinuxPermissiveDecision(
                set_permissive=True))
            report_generic(
                title='SElinux will be set to permissive mode',
                summary='SElinux will be set to permissive mode as it was in enforcing mode.',
                severity='low')



from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api, run


def is_config_default():
    """Check if the chrony config file was not modified since installation."""
    try:
        result = run(['rpm', '-V', '--nomtime', 'chrony'], checked=False)
        return '/etc/chrony.conf' not in result['stdout']
    except OSError as e:
        api.current_logger().warn("rpm verification failed: %s", str(e))
        return True


def check_chrony(chrony_installed):
    """Report potential issues in chrony configuration."""
    if not chrony_installed:
        api.current_logger().info('chrony package is not installed')
        return

    if is_config_default():
        reporting.report_generic(
            title='chrony using default configuration',
            summary='default chrony configuration in RHEL8 uses leapsectz directive, which cannot be used with '
            'leap smearing NTP servers, and uses a single pool directive instead of four server directives',
            severity='medium')
    else:
        reporting.report_generic(
            title='chrony using non-default configuration',
            summary='chrony behavior will not change in RHEL8',
            severity='low')



from leapp.snactor.fixture import current_actor_context
from leapp.models import RPM, InstalledRedHatSignedRPM, FilteredRpmTransactionTasks, RpmTransactionTasks


RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(FilteredRpmTransactionTasks)


def test_actor_execution_with_sample_data(current_actor_context):
    installed_rpm = [
        RPM(name='sample01', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_PGP_SIG'),
        RPM(name='sample02', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_PGP_SIG')]
    current_actor_context.feed(InstalledRedHatSignedRPM(items=installed_rpm))
    current_actor_context.feed(RpmTransactionTasks(
        to_remove=[rpm.name for rpm in installed_rpm],
        to_keep=[installed_rpm[0].name]
    ))
    current_actor_context.run()
    result = current_actor_context.consume(FilteredRpmTransactionTasks)
    assert len(result) == 1
    assert result[0].to_keep == [installed_rpm[0].name]
    assert result[0].to_remove == [installed_rpm[1].name]



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_with_remediation
from leapp.libraries.common.rpms import has_package
from leapp.models import InstalledRedHatSignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckGrep(Actor):
    """
    Check if Grep is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkgrep'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(InstalledRedHatSignedRPM, 'grep'):
            report_with_remediation(
                title='Grep has incompatible changes in the next major version',
                summary='If a file contains data improperly encoded for the current locale, and this is '
                        'discovered before any of the file\'s contents are output, grep now treats the file '
                        'as binary.\n'
                        'The \'grep -P\' no longer reports an error and exits when given invalid UTF-8 data. '
                        'Instead, it considers the data to be non-matching.\n'
                        'In locales with multibyte character encodings other than UTF-8, grep -P now reports '
                        'an error and exits instead of misbehaving.\n'
                        'When searching binary data, grep now may treat non-text bytes as line terminators. '
                        'This can boost performance significantly.\n'
                        'The \'grep -z\' no longer automatically treats the byte \'\\200\' as binary data.\n'
                        'Context no longer excludes selected lines omitted because of -m. For example, '
                        '\'grep "^" -m1 -A1\' now outputs the first two input lines, not just the first '
                        'line.\n',
                remediation='Please update your scripts to be compatible with the changes.',
                severity='low')





def semantics_changes(config):
    globally_enabled = False
    in_match_disabled = False
    for opt in config.permit_root_login:
        if opt.value != "yes" and opt.in_match is not None \
                and opt.in_match[0].lower() != 'all':
            in_match_disabled = True

        if opt.value == "yes" and (opt.in_match is None or
                                   opt.in_match[0].lower() == 'all'):
            globally_enabled = True

    return not globally_enabled and in_match_disabled



from leapp.models import LeftoverPackages
from leapp.snactor.fixture import current_actor_context


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(LeftoverPackages)



from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import UsedTargetRepositories


CRB_REPOID = 'codeready-builder-for-rhel-8-x86_64-rpms'


def _is_crb_used():
    # the UsedTargetRepositories has to be set always, by design of IPU
    used_repos = next(api.consume(UsedTargetRepositories), None)
    for repo in used_repos.repos:
        if repo.repoid == CRB_REPOID:
            return True
    return False


def process():
    if _is_crb_used():
        try:
            run(['subscription-manager', 'repos', '--enable', CRB_REPOID])
        except CalledProcessError as e:
            api.current_logger().error('Cannot enable CRB repository: %s', str(e))



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api
from leapp.libraries.actor import library
from leapp.models import BootContent


class run_mocked(object):
    args = []

    def __call__(self, args, split=True):
        self.args.append(args)


def test_remove_boot_entry(monkeypatch):
    def get_upgrade_kernel_filepath_mocked():
        return '/abc'
    monkeypatch.setattr(library, 'get_upgrade_kernel_filepath', get_upgrade_kernel_filepath_mocked)
    monkeypatch.setattr(library, 'run', run_mocked())

    library.remove_boot_entry()

    assert library.run.args == [['/bin/mount', '/boot'],
                                ['/usr/sbin/grubby', '--remove-kernel=/abc'], ['/bin/mount', '-a']]


def test_get_upgrade_kernel_filepath(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/abc', initram_path='/def')
    monkeypatch.setattr(api, 'consume', consume_message_mocked)

    kernel_path = library.get_upgrade_kernel_filepath()

    assert kernel_path == '/abc'

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None
    monkeypatch.setattr(api, 'consume', consume_no_message_mocked)

    with pytest.raises(StopActorExecutionError):
        library.get_upgrade_kernel_filepath()



from leapp.actors import Actor
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag
from leapp.models import SelinuxRelabelDecision
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_generic, report_with_remediation


class ScheduleSeLinuxRelabeling(Actor):
    """
    Schedule SELinux relabelling.

    If SELinux status was set to permissive or enforcing, a relabelling is necessary.
    """

    name = 'schedule_se_linux_relabelling'
    consumes = (SelinuxRelabelDecision,)
    produces = (Report,)
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        for decision in self.consume(SelinuxRelabelDecision):
            if decision.set_relabel:
                try:
                    with open('/.autorelabel', 'w'):
                        pass
                    report_generic(
                        title='SElinux scheduled for relabelling',
                        summary='/.autorelabel file touched on root in order to schedule SElinux relabelling.',
                        severity='low',
                    )
                except OSError as e:
                    # FIXME: add an "action required" flag later
                    report_with_remediation(
                        title='Could not schedule SElinux for relabelling',
                        summary='./autorelabel file could not be created: {}.'.format(e),
                        remediation='Please set autorelabelling manually after the upgrade.',
                        severity='high'
                    )
                    self.log.critical('Could not schedule SElinux for relabelling: %s.' % e)



import pytest

from leapp.exceptions import StopActorExecution
from leapp.libraries.stdlib import api
from leapp.libraries.actor import library
from leapp.models import BootContent


class remove_file_mocked(object):
    def __init__(self):
        self.called = 0
        self.files_to_remove = []

    def __call__(self, filename):
        self.called += 1
        self.files_to_remove.append(filename)


class logger_mocked(object):
    def __init__(self):
        self.warnmsg = None
        self.errmsg = None

    def warning(self, msg):
        self.warnmsg = msg

    def error(self, msg):
        self.errmsg = msg

    def __call__(self):
        return self


def test_remove_boot_files(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/abc', initram_path='/def')
    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_message_mocked)
    monkeypatch.setattr(library, 'remove_file', remove_file_mocked())

    library.remove_boot_files()

    assert library.remove_file.files_to_remove == ['/abc', '/def']

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None
    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_no_message_mocked)
    monkeypatch.setattr(library, 'remove_file', remove_file_mocked())
    monkeypatch.setattr('leapp.libraries.stdlib.api.current_logger', logger_mocked())

    with pytest.raises(StopActorExecution):
        library.remove_boot_files()

    assert library.remove_file.called == 0
    assert "Did not receive a message" in api.current_logger.warnmsg


def test_remove_file_that_does_not_exist(monkeypatch):
    def remove_mocked(filepath):
        raise OSError
    monkeypatch.setattr('os.remove', remove_mocked)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    library.remove_file('/filepath')

    assert "Could not remove /filepath" in api.current_logger.errmsg



import json


class PlainTextReport(object):
    def header(self, report_file):
        pass

    def write(self, result, report_file):
        report_file.write('Severity: ' + result.severity + '\n')
        report_file.write('Title: ' + result.title + '\n')
        report_file.write('Summary: ' + result.detail['summary'] + '\n')
        if result.detail.get('remediation'):
            report_file.write('Remediation: ' + result.detail['remediation'] + '\n')
        report_file.write('-' * 40 + '\n')

    def footer(self, report_file):
        pass


class JSONReport(object):
    _first_entry = True

    def header(self, report_file):
        report_file.write('[')

    def write(self, result, report_file):

        data = json.dumps(result.dump())

        if self._first_entry:
            report_file.write(data)
            self._first_entry = False
        else:
            report_file.write(',' + data)

    def footer(self, report_file):
        report_file.write(']\n')


def generate_report(results, path):
    reporter = None

    if path.endswith('.txt'):
        reporter = PlainTextReport()

    if path.endswith('.json'):
        reporter = JSONReport()

    if not reporter:
        return 'Report format not supported.'

    with open(path, 'w') as report_file:
        reporter.header(report_file)

        for r in results:
            reporter.write(r, report_file)

        reporter.footer(report_file)

    return None



import os
import errno

from leapp.snactor.fixture import current_actor_context


def test_remove_resume_service(current_actor_context):
    service_name = 'leapp_resume.service'
    service_path = os.path.join('/etc/systemd/system/', service_name)
    symlink_path = os.path.join('/etc/systemd/system/default.target.wants/', service_name)

    # lets make sure there are not leftovers from previous tests
    try:
        os.unlink(service_path)
        os.unlink(symlink_path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

    with open(service_path, 'w'):
        pass

    os.symlink(service_path, symlink_path)

    current_actor_context.run()

    assert not os.path.isfile(service_path)
    assert not os.path.isfile(symlink_path)



import os.path

from leapp.libraries.common.reporting import report_generic

COROSYNC_CONF_LOCATION = "/etc/corosync/corosync.conf"
CIB_LOCATION = "/var/lib/pacemaker/cib/cib.xml"


def inhibit(node_type):
    report_generic(
        title="Use of HA cluster detected. Upgrade can't proceed.",
        summary=(
            "HA cluster is not supported by the inplace upgrade.\n"
            "HA cluster configuration file(s) found."
            " It seems to be a cluster {0}.".format(node_type)
        ),
        severity="high",
        flags=["inhibitor"],
    )


def check_ha_cluster():
    if os.path.isfile(COROSYNC_CONF_LOCATION):
        inhibit(node_type="node")
    elif os.path.isfile(CIB_LOCATION):
        inhibit(node_type="remote node")



from leapp.actors import Actor
from leapp.libraries.common import dnfplugin
from leapp.models import FilteredRpmTransactionTasks, TargetUserSpaceInfo, UsedTargetRepositories, XFSPresence
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class DnfTransactionCheck(Actor):
    """
    This actor tries to solve the RPM transaction to verify the all package dependencies can be successfully resolved.
    """

    name = 'dnf_transaction_check'
    consumes = (UsedTargetRepositories, FilteredRpmTransactionTasks, TargetUserSpaceInfo, XFSPresence)
    produces = ()
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):

        presence = next(self.consume(XFSPresence), XFSPresence())
        xfs_present = presence.present and presence.without_ftype
        used_repos = self.consume(UsedTargetRepositories)
        tasks = next(self.consume(FilteredRpmTransactionTasks), FilteredRpmTransactionTasks())
        target_userspace_info = next(self.consume(TargetUserSpaceInfo), None)

        dnfplugin.perform_transaction_check(tasks=tasks, used_repos=used_repos,
                                            target_userspace_info=target_userspace_info, xfs_present=xfs_present)



from leapp.models import Model, fields
from leapp.topics import RHSMTopic


class RHSMInfo(Model):
    """
    Subscription manager details required for the inplace upgrade.
    """
    topic = RHSMTopic

    release = fields.Nullable(fields.String())
    """ Release the subscription manager is set to. """
    attached_skus = fields.List(fields.String(), default=[])
    """ SKUs the current system is attached to. """
    available_repos = fields.List(fields.String(), default=[])
    """ Repositories that are available to the current system through the subscription manager. """
    enabled_repos = fields.List(fields.String(), default=[])
    """ Repositories that are enabled on the current system through the subscription manager. """
    existing_product_certificates = fields.List(fields.String(), default=[])
    """ Product certificates that are currently installed on the system. """


class SourceRHSMInfo(RHSMInfo):
    pass


class TargetRHSMInfo(RHSMInfo):
    pass



from leapp.models import Model, fields
from leapp.topics import TargetUserspaceTopic


class RequiredTargetUserspacePackages(Model):
    topic = TargetUserspaceTopic
    packages = fields.List(fields.String(), default=[])



from leapp.actors import Actor
from leapp.libraries.common import reporting
from leapp.tags import ChecksPhaseTag
from leapp.models import Report


class PythonInformUser(Actor):
    name = "python_inform_user"
    description = "This actor informs the user of differences in Python version and support in RHEL 8."
    consumes = ()
    produces = (Report,)
    tags = (ChecksPhaseTag,)

    def process(self):
        url = "https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/8/html-single/configuring_basic_system_settings/#using-python3"  # noqa: E501; pylint: disable=line-too-long
        title = "Difference in Python versions and support in RHEL 8"
        severity = "high"
        summary = ("In RHEL 8, there is no 'python' command."
                   " Python 3 (backward incompatible) is the primary Python version"
                   " and Python 2 is available with limited support and limited set of packages."
                   " Read more here: {}".format(url))
        remediation = "alternatives --set python /usr/bin/python3"

        reporting.report_with_remediation(title=title,
                                          severity=severity,
                                          summary=summary,
                                          remediation=remediation)



from leapp.actors import Actor
from leapp.libraries.actor.scanner import detect_config_error
from leapp.models import GrubConfigError
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_generic
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class DetectGrubConfigError(Actor):
    """
    Check grub configuration for syntax error in GRUB_CMDLINE_LINUX value.
    """

    name = 'detect_grub_config_error'
    consumes = ()
    produces = (Report, GrubConfigError)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        error_detected = detect_config_error('/etc/default/grub')
        if error_detected:
            report_generic(
                title='Syntax error detected in grub configuration',
                summary='Syntax error was detected in GRUB_CMDLINE_LINUX value of grub configuration. '
                        'This error is causing booting and other issues. '
                        'Error is automatically fixed by add_upgrade_boot_entry actor.',
                severity='low'
            )

        self.produce(GrubConfigError(error_detected=error_detected))



import platform

from leapp.actors import Actor
from leapp.models import (CustomTargetRepository, RepositoriesBlacklisted, RepositoriesFacts, RepositoriesMap,
                          RepositoriesSetupTasks, RHELTargetRepository, SkippedRepositories, TargetRepositories,
                          UsedRepositories)
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SetupTargetRepos(Actor):
    """
    Produces list of repositories that should be available to be used by Upgrade process.

    Based on current set of Red Hat Enterprise Linux repositories, produces the list of target
    repositories. Additionaly process request to use custom repositories during the upgrade
    transaction.
    """

    name = 'setuptargetrepos'
    consumes = (CustomTargetRepository,
                RepositoriesSetupTasks,
                RepositoriesMap,
                RepositoriesFacts,
                RepositoriesBlacklisted,
                UsedRepositories)
    produces = (TargetRepositories, SkippedRepositories)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        # TODO: Think about Beta and Alpha repositories. How will we tell we
        # + want to go to GA, Alpha, Beta, ... repos?

        custom_repos = []
        for repo in self.consume(CustomTargetRepository):
            custom_repos.append(repo)

        enabled_repos = set()
        for repos in self.consume(RepositoriesFacts):
            for repo_file in repos.repositories:
                for repo in repo_file.data:
                    if repo.enabled:
                        enabled_repos.add(repo.repoid)

        rhel_repos = []
        mapped_repos = set()
        for repos_map in self.consume(RepositoriesMap):
            for repo_map in repos_map.repositories:
                # Check if repository map architecture matches system architecture
                if platform.machine() != repo_map.arch:
                    continue

                mapped_repos.add(repo_map.from_id)
                if repo_map.from_id in enabled_repos:
                    rhel_repos.append(RHELTargetRepository(repoid=repo_map.to_id))

        skipped_repos = enabled_repos.difference(mapped_repos)

        used = {}
        for used_repos in self.consume(UsedRepositories):
            for used_repo in used_repos.repositories:
                used[used_repo.repository] = used_repo.packages
                for repo in repo_file.data:
                    enabled_repos.add(repo.repoid)

        skipped_repos = skipped_repos.intersection(set(used.keys()))

        if skipped_repos:
            pkgs = set()
            for repo in skipped_repos:
                pkgs.update(used[repo])
            self.produce(SkippedRepositories(repos=list(skipped_repos), packages=list(pkgs)))

        for task in self.consume(RepositoriesSetupTasks):
            for repo in task.to_enable:
                rhel_repos.append(RHELTargetRepository(repoid=repo))

        repos_blacklisted = set()
        for blacklist in self.consume(RepositoriesBlacklisted):
            repos_blacklisted.update(blacklist.repoids)
        rhel_repos = [repo for repo in rhel_repos if repo.repoid not in repos_blacklisted]
        custom_repos = [repo for repo in custom_repos if repo.repoid not in repos_blacklisted]

        self.produce(TargetRepositories(
            rhel_repos=rhel_repos,
            custom_repos=custom_repos,
        ))

        # TODO: Some informational messages would be added for the report and
        # + logs, so we and user will know exactly what is going on.

        return



import contextlib
import functools
import os
import re

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import TargetRHSMInfo


_RE_REPO_UID = re.compile(r'Repo ID:\s*([^\s]+)')
_RE_RELEASE = re.compile(r'Release:\s*([^\s]+)')
_RE_SKU_CONSUMED = re.compile(r'SKU:\s*([^\s]+)')


@contextlib.contextmanager
def _handle_rhsm_exceptions(hint=None):
    """
    Context manager based function that handles exceptions of `run` for the subscription manager calls.
    """
    try:
        yield
    except OSError as e:
        api.current_logger().error('Failed to execute subscription-manager executable')
        raise StopActorExecutionError(
            message='Unable to execute subscription-manager executable. Message: {}'.format(str(e)),
            details={
                'hint': 'Please ensure subscription-manager is installed and exceutable.'
            }
        )
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='A subscription-manager command failed to execute',
            details={
                'hint': hint or 'Please ensure you have a valid RHEL subscription and your network is up.'
            }
        )


def skip_rhsm():
    """ Function to check whether we should skip RHSM related code """
    return os.getenv('LEAPP_DEVEL_SKIP_RHSM', '0') == '1'


def with_rhsm(f):
    """ Decorator to allow skipping RHSM functions by executing a no-op """
    if skip_rhsm():
        @functools.wraps(f)
        def _no_op(*args, **kwargs):
            return
        return _no_op
    return f


@with_rhsm
def get_attached_skus(context, rhsm_info):
    """
    Retrieves the list of SKU the system is attached to with the subscription manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if not rhsm_info.attached_skus:
        with _handle_rhsm_exceptions():
            result = context.call(['subscription-manager', 'list', '--consumed'], split=False)
            rhsm_info.attached_skus = _RE_SKU_CONSUMED.findall(result['stdout'])


@with_rhsm
def get_available_repo_uids(context, rhsm_info):
    """
    Retrieves all available repositories names from the subscription manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if not rhsm_info.available_repos:
        with _handle_rhsm_exceptions():
            result = context.call(['subscription-manager', 'repos'], split=False)
            rhsm_info.available_repos = _RE_REPO_UID.findall(result['stdout'])


@with_rhsm
def get_enabled_repo_uids(context, rhsm_info):
    """
    Retrieves all enabled repositories names from the subscription manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if not rhsm_info.enabled_repos:
        with _handle_rhsm_exceptions():
            result = context.call(['subscription-manager', 'repos', '--list-enabled'], split=False)
            rhsm_info.enabled_repos = _RE_REPO_UID.findall(result['stdout'])


@with_rhsm
def _get_repositories_to_use(context, rhsm_info, target_repositories):
    """
    Filters the available repositories based on the target_repositories passed.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    :param target_repositories: Instance of the TargetRepositories message.
    :type target_repositories: TargetRepositories
    :return: List of filtered repositories to use.
    :rtype: List(string)
    """
    repositories_to_use = []
    for target_repo in target_repositories:
        for rhel_repo in target_repo.rhel_repos:
            if rhel_repo.uid in rhsm_info.available_repos:
                repositories_to_use.append(rhel_repo.uid)
    return repositories_to_use


@with_rhsm
def unset_release(context):
    """
    Unsets the configured release from the subscription manager so we can perform the upgrade.
    Stores the previous set release in self.info if not already received.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'release', '--unset'], split=False)


@with_rhsm
def set_release(context, release):
    """
    This function will set the version specified.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param release: Release to set the subscription manager to.
    :type release: str
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'release', '--set', release], split=False)


@with_rhsm
def restore_release(context, rhsm_info):
    """
    If a release has been set, this function will restore it from the rhsm_info.release value.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if rhsm_info.release:
        set_release(context, rhsm_info.release)


@with_rhsm
def get_release(context, rhsm_info):
    """
    Retrieves the release the subscription manager has been pinned to, if applicable.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if rhsm_info.release is None:
        with _handle_rhsm_exceptions():
            result = context.call(['subscription-manager', 'release'], split=False)
            result = _RE_RELEASE.findall(result['stdout'])
            rhsm_info.release = result[0] if result else ''


@with_rhsm
def refresh(context):
    """
    Calls 'subscription-manager refresh'

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'refresh'], split=False)


@with_rhsm
def get_existing_product_certificates(context, rhsm_info):
    """
    Retrieves information about existing product certificates on the system.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    if not rhsm_info.existing_product_certificates:
        for path in ('/etc/pki/product', '/etc/pki/product-default'):
            if not os.path.isdir(context.full_path(path)):
                continue
            certs = [os.path.join(path, f) for f in os.listdir(context.full_path(path))
                     if os.path.isfile(os.path.join(context.full_path(path), f))]
            if not certs:
                continue
            rhsm_info.existing_product_certificates.extend(certs)


@contextlib.contextmanager
def switched_certificate(context, rhsm_info, cert_path, version):
    """
    Performs all actions needed to switch the product certificate passed.

    This function will copy the certificate to /etc/pki/product and if necessary /etc/pki/product-default and
    removes other product certificates from there. Unsets the release and refreshes the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    :param cert_path: Path to the product certificate to switch to.
    :type cert_path: string
    """
    if skip_rhsm():
        yield TargetRHSMInfo()
        return

    # Make a backup of product certificates
    pki_path = '/etc/pki'
    pki_backup_path = '/etc/pki.bak'
    context.call(['rm', '-rf', pki_backup_path], checked=False)
    context.call(['cp', '-a', pki_path, pki_backup_path], checked=False)

    for existing in rhsm_info.existing_product_certificates:
        try:
            context.remove(existing)
        except OSError:
            api.current_logger().warn('Failed to remove existing certificate: %s', existing, exc_info=True)

    for path in ('/etc/pki/product', '/etc/pki/product-default'):
        if os.path.isdir(context.full_path(path)):
            context.copy_to(cert_path, os.path.join(path, os.path.basename(cert_path)))

    unset_release(context)
    try:
        refresh(context)
        set_release(context, version)
        target_rhsm_info = TargetRHSMInfo()
        scan_rhsm_info(context, target_rhsm_info)
        yield target_rhsm_info
    finally:
        # Restore backup of product certificates
        context.call(['rm', '-rf', pki_path], checked=False)
        context.call(['cp', '-a', pki_backup_path, pki_path], checked=False)
        unset_release(context)
        # Restore release
        restore_release(context, rhsm_info)


@with_rhsm
def scan_rhsm_info(context, rhsm_info):
    """
    Gathers all RHSM information

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of a RHSMInfo derived model.
    :type rhsm_info: RHSMInfo derived model
    """
    get_attached_skus(context, rhsm_info)
    get_available_repo_uids(context, rhsm_info)
    get_enabled_repo_uids(context, rhsm_info)
    get_release(context, rhsm_info)
    get_existing_product_certificates(context, rhsm_info)



from leapp.snactor.fixture import current_actor_context
from leapp.models import InstalledRPM


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(InstalledRPM)
    assert current_actor_context.consume(InstalledRPM)[0].items



from leapp.libraries import stdlib
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelVersion


def process():
    kernels = stdlib.run(["rpm", "-q", "kernel"], split=True)["stdout"]
    for kernel in kernels:
        version = kernel.split("-", 1)[1]
        if "el8" in version:
            api.produce(InstalledTargetKernelVersion(version=version))
            break



import re


def detect_config_error(conf_file):
    '''
    Check grub configuration for syntax error in GRUB_CMDLINE_LINUX value.

    :return: Function returns True if error was detected, otherwise False.
    '''
    with open(conf_file, 'r') as f:
        config = f.read()

    pattern = r'GRUB_CMDLINE_LINUX="[^"]+"(?!(\s*$)|(\s+GRUB))'
    return re.search(pattern, config) is not None



from leapp.actors import Actor
from leapp.models import StorageInfo
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_with_remediation
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckNfs(Actor):
    """
    Check if NFS filesystem is in use. If yes, inhibit the upgrade process.

    Actor looks for NFS in the following sources: /ets/fstab, mount and systemd-mount.
    If there is NFS in any of the mentioned sources, actors inhibits the upgrade.
    """
    name = "check_nfs"
    consumes = (StorageInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag,)

    def process(self):
        details = "NFS is currently not supported by the inplace upgrade.\n" \
                  "We have found NFS usage at the following locations:\n"
        nfs_found = False

        for storage in self.consume(StorageInfo):
            # Check fstab
            for fstab in storage.fstab:
                if fstab.fs_vfstype == "nfs":
                    nfs_found = True
                    details += "- One or more NFS entries in /etc/fstab\n"
                    break

            # Check mount
            for mount in storage.mount:
                if mount.tp == "nfs":
                    nfs_found = True
                    details += "- Currently mounted NFS shares\n"
                    break

            # Check systemd-mount
            for systemdmount in storage.systemdmount:
                if systemdmount.fs_type == "nfs":
                    nfs_found = True
                    details += "- One or more configured NFS mounts in systemd-mount\n"
                    break

        if nfs_found:
            report_with_remediation(title="Use of NFS detected. Upgrade can't proceed",
                                    summary=details,
                                    remediation="Disable NFS temporarily for the upgrade if possible.",
                                    severity='high',
                                    flags=['inhibitor'])



import itertools
import os

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import constants
from leapp.libraries.common import dnfplugin, mounting, overlaygen, rhsm, utils
from leapp.libraries.stdlib import api, run
from leapp.models import (OSReleaseFacts, RequiredTargetUserspacePackages, SourceRHSMInfo, TargetRepositories,
                          TargetUserSpaceInfo, UsedTargetRepositories, UsedTargetRepository, XFSPresence)


def prepare_target_userspace(context, userspace_dir, enabled_repos, packages):
    """
    Implements the creation of the target userspace.
    """
    run(['rm', '-rf', userspace_dir])
    _create_target_userspace_directories(userspace_dir)
    with mounting.BindMount(source=userspace_dir, target=os.path.join(context.base_dir, 'el8target')):
        repos_opt = [['--enablerepo', repo] for repo in enabled_repos]
        repos_opt = list(itertools.chain(*repos_opt))
        context.call(
            [
                'dnf',
                'install',
                '-y',
                '--nogpgcheck',
                '--setopt=module_platform_id=platform:el8',
                '--setopt=keepcache=1',
                '--releasever', '8',
                '--installroot', '/el8target',
                '--disablerepo', '*'
            ] + repos_opt + packages,
            callback_raw=utils.logging_handler
        )


def _prep_repository_access(context, target_userspace):
    """
    Prepare repository access by copying all relevant certificates and configuration files to the userspace
    """
    if not rhsm.skip_rhsm():
        run(['rm', '-rf', os.path.join(target_userspace, 'etc', 'pki')])
        run(['rm', '-rf', os.path.join(target_userspace, 'etc', 'rhsm')])
        context.copytree_from('/etc/pki', os.path.join(target_userspace, 'etc', 'pki'))
        context.copytree_from('/etc/rhsm', os.path.join(target_userspace, 'etc', 'rhsm'))
    run(['rm', '-rf', os.path.join(target_userspace, 'etc', 'yum.repos.d')])
    context.copytree_from('/etc/yum.repos.d', os.path.join(target_userspace, 'etc', 'yum.repos.d'))


def get_os_variant():
    """
    Retrieves the OS variant from the first matching OSReleaseFacts message. Returns an empty string otherwise.
    """
    for msg in api.consume(OSReleaseFacts):
        if msg.variant_id:
            return msg.variant_id
    return ''


def _get_product_certificate_path():
    """
    Retrieves the required / used product certificate for RHSM.
    """
    sys_var = get_os_variant()
    var_prodcert = {'server': '479.pem'}
    if sys_var not in var_prodcert:
        raise StopActorExecutionError(
            message=('Failed to to retrieve Product Cert file.'
                     'Product cert file not available for System Variant \'{}\'.'.format(sys_var))
        )

    return api.get_file_path(var_prodcert[sys_var])


def _create_target_userspace_directories(target_userspace):
    api.current_logger().debug('Creating target userspace directories.')
    try:
        utils.makedirs(target_userspace)
        api.current_logger().debug('Done creating target userspace directories.')
    except OSError:
        api.current_logger().error(
            'Failed to create temporary target userspace directories %s', target_userspace, exc_info=True)
        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {directory} is empty and modifiable.'.format(directory=target_userspace)
            }
        )


def gather_target_repositories(target_rhsm_info):
    """
    Performs basic checks on requirements for RHSM repositories and returns the list of target repository ids to use
    during the upgrade.
    """
    # FIXME: check that required repo IDs (baseos, appstream)
    # + or check that all required RHEL repo IDs are available.
    if not rhsm.skip_rhsm():
        if not target_rhsm_info.available_repos or len(target_rhsm_info.available_repos) < 2:
            raise StopActorExecutionError(
                message='Cannot find required basic RHEL repositories.',
                details={
                    'hint': ('It is required to have RHEL repository on the system'
                             ' provided by the subscription-manager. Possibly you'
                             ' are missing a valid SKU for the target system or network'
                             ' connection failed. Check whether your system is attached'
                             ' to the valid SKU providing target repositories.')
                }
            )

    target_repoids = []
    for target_repo in api.consume(TargetRepositories):
        for rhel_repo in target_repo.rhel_repos:
            if rhel_repo.repoid in target_rhsm_info.available_repos:
                target_repoids.append(rhel_repo.repoid)
        for custom_repo in target_repo.custom_repos:
            # TODO: complete processing of custom repositories
            # HINT: now it will work only for custom repos that exist
            # + already on the system in a repo file
            # TODO: should check available_target_repoids + additional custom repos
            # + outside of rhsm..
            # #if custom_repo.repoid in available_target_repoids:
            target_repoids.append(custom_repo.repoid)
    return target_repoids


def perform():
    packages = {'dnf'}
    for message in api.consume(RequiredTargetUserspacePackages):
        packages.update(message.packages)

    rhsm_info = next(api.consume(SourceRHSMInfo), None)
    if not rhsm_info and not rhsm.skip_rhsm():
        api.current_logger().warn("Could not receive RHSM information - Is this system registered?")
        return

    presence = next(api.consume(XFSPresence), XFSPresence())
    xfs_present = presence.present and presence.without_ftype

    prod_cert_path = _get_product_certificate_path()
    with overlaygen.create_source_overlay(
            mounts_dir=constants.MOUNTS_DIR,
            scratch_dir=constants.SCRATCH_DIR,
            xfs_present=xfs_present) as overlay:
        with overlay.nspawn() as context:
            target_version = api.current_actor().configuration.version.target
            with rhsm.switched_certificate(context, rhsm_info, prod_cert_path, target_version) as target_rhsm_info:
                api.current_logger().debug("Target RHSM Info: SKUs: {skus} Repositories: {repos}".format(
                    repos=target_rhsm_info.enabled_repos,
                    skus=rhsm_info.attached_skus if rhsm_info else []
                ))
                target_repoids = gather_target_repositories(target_rhsm_info)
                api.current_logger().debug("Gathered target repositories: {}".format(', '.join(target_repoids)))
                if not target_repoids:
                    raise StopActorExecutionError(
                        message='There are no enabled target repositories for the upgrade process to proceed.',
                        details={'hint': (
                            'Ensure your system is correctly registered with the subscription manager and that'
                            ' your current subscription is entitled to install the requested target version {version}'
                            ).format(version=api.current_actor().configuration.version.target)
                        }
                    )
                prepare_target_userspace(context, constants.TARGET_USERSPACE, target_repoids, list(packages))
                _prep_repository_access(context, constants.TARGET_USERSPACE)
                dnfplugin.install(constants.TARGET_USERSPACE)
                api.produce(UsedTargetRepositories(
                    repos=[UsedTargetRepository(repoid=repo) for repo in target_repoids]))
                api.produce(target_rhsm_info)
                api.produce(TargetUserSpaceInfo(
                    path=constants.TARGET_USERSPACE,
                    scratch=constants.SCRATCH_DIR,
                    mounts=constants.MOUNTS_DIR))



from leapp.actors import Actor
from leapp.models import Report, OpenSshConfig
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api
from leapp.libraries.common.reporting import report_generic


class OpenSshUsePrivilegeSeparationCheck(Actor):
    """
    UsePrivilegeSeparation configuration option was removed.

    Check the value of UsePrivilegeSeparation in OpenSSH server config file
    and warn about its deprecation if it is set to non-default value.
    """
    name = 'open_ssh_use_privilege_separation'
    consumes = (OpenSshConfig, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        openssh_messages = self.consume(OpenSshConfig)
        config = next(openssh_messages)
        if list(openssh_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')

        if config.use_privilege_separation is not None and \
           config.use_privilege_separation != "sandbox":
            report_generic(
                title='OpenSSH configured not to use privilege separation sandbox',
                summary='OpenSSH is configured to disable privilege '
                        'separation sandbox, which is decreasing security '
                        'and is no longer supported in RHEL 8',
                severity='low')



import csv
import os

from leapp.exceptions import StopActorExecution
from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api
from leapp.models import RepositoriesMap, RepositoryMap
from leapp.models.fields import ModelViolationError


def inhibit_upgrade(title):
    summary = 'Read documentation at: https://access.redhat.com/articles/3664871 for more information ' \
        'about how to retrieve the files'
    reporting.report_generic(title=title, summary=summary, severity='high', flags=['inhibitor'])
    raise StopActorExecution()


def scan_repositories(path):
    if not os.path.isfile(path):
        inhibit_upgrade('Repositories map file not found ({})'.format(path))

    if os.path.getsize(path) == 0:
        inhibit_upgrade('Repositories map file is invalid ({})'.format(path))

    repositories = []
    with open(path) as f:
        data = csv.reader(f)
        next(data)  # skip header

        for row in data:
            # skip empty lines and comments
            if not row or row[0].startswith('#'):
                continue

            try:
                from_id, to_id, from_minor_version, to_minor_version, arch, repo_type = row
            except ValueError as err:
                inhibit_upgrade('Repositories map file is invalid, offending line number: {} ({})'.format(
                    data.line_num, err))

            try:
                repositories.append(RepositoryMap(from_id=from_id,
                                                  to_id=to_id,
                                                  from_minor_version=from_minor_version,
                                                  to_minor_version=to_minor_version,
                                                  arch=arch,
                                                  repo_type=repo_type))
            except ModelViolationError as err:
                inhibit_upgrade('Repositories map file is invalid, offending line number: {} ({})'.format(
                    data.line_num, err))

    if not repositories:
        inhibit_upgrade('Repositories map file is invalid ({})'.format(path))

    api.produce(RepositoriesMap(repositories=repositories))



from leapp.libraries.actor.scanner import detect_config_error


def test_correct_config():
    assert not detect_config_error('files/grub.correct')
    assert not detect_config_error('files/grub.correct_trailing_space')


def test_wrong_config():
    assert detect_config_error('files/grub.wrong')
    assert detect_config_error('files/grub.wrong1')



from leapp.actors import Actor
from leapp.models import Report, OpenSshConfig
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api
from leapp.libraries.common.reporting import report_with_remediation


class OpenSshAlgorithmsCheck(Actor):
    """
    OpenSSH configuration does not contain any unsupported cryptographic algorithms.

    Check the values of Ciphers and MACs in OpenSSH server config file and warn
    about removed algorithms which might cause the server to fail to start.
    """
    name = 'open_ssh_algorithms'
    consumes = (OpenSshConfig,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        removed_ciphers = [
            "blowfish-cbc",
            "cast128-cbc",
            "arcfour",
            "arcfour128",
            "arcfour256",
        ]
        removed_macs = [
            "hmac-ripemd160",
        ]
        found_ciphers = []
        found_macs = []
        openssh_messages = self.consume(OpenSshConfig)
        config = next(openssh_messages)
        if list(openssh_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')

        for cipher in removed_ciphers:
            if config.ciphers and cipher in config.ciphers:
                found_ciphers.append(cipher)
        for mac in removed_macs:
            if config.macs and mac in config.macs:
                found_macs.append(mac)

        if found_ciphers:
            report_with_remediation(
                title='OpenSSH configured to use removed ciphers',
                summary='OpenSSH is configured to use removed ciphers {}. '
                        'These ciphers were removed from OpenSSH and if '
                        'present the sshd daemon will not start in RHEL 8'
                        ''.format(','.join(found_ciphers)),
                remediation='Remove the following ciphers from sshd_config: '
                            '{}'.format(','.join(found_ciphers)),
                severity='high',
                flags=['inhibitor'])

        if found_macs:
            report_with_remediation(
                title='OpenSSH configured to use removed mac',
                summary='OpenSSH is configured to use removed mac {}. '
                        'This MAC was removed from OpenSSH and if present '
                        'the sshd daemon will not start in RHEL 8'
                        ''.format(','.join(found_macs)),
                remediation='Remove the following MACs from sshd_config: '
                            '{}'.format(','.join(found_macs)),
                severity='high',
                flags=['inhibitor'])



from leapp.actors import Actor
from leapp.models import InstalledRedHatSignedRPM, InstalledUnsignedRPM, InstalledRPM
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class RedHatSignedRpmScanner(Actor):
    """
    Provides data about installed RPM Packages signed by Red Hat.

    After filtering the list of installed RPM packages by signature, a message with relevant data
    will be produced.
    """

    name = 'red_hat_signed_rpm_scanner'
    consumes = (InstalledRPM,)
    produces = (InstalledRedHatSignedRPM, InstalledUnsignedRPM,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        RH_SIGS = ['199e2f91fd431d51',
                   '5326810137017186',
                   '938a80caf21541eb',
                   'fd372689897da07a',
                   '45689c882fa658e0']

        signed_pkgs = InstalledRedHatSignedRPM()
        unsigned_pkgs = InstalledUnsignedRPM()

        for rpm_pkgs in self.consume(InstalledRPM):
            for pkg in rpm_pkgs.items:
                # "gpg-pubkey" is not signed as it would require another package to verify its signature
                if any(key in pkg.pgpsig for key in RH_SIGS) or \
                        (pkg.name == 'gpg-pubkey' and pkg.packager.startswith('Red Hat, Inc.')):
                    signed_pkgs.items.append(pkg)
                    continue

                unsigned_pkgs.items.append(pkg)

        self.produce(signed_pkgs)
        self.produce(unsigned_pkgs)
